# Generated by the Protocol Buffers compiler. DO NOT EDIT!
# source: pymap/admin/grpc/admin.proto
# plugin: grpclib.plugin.main
import abc

import grpclib.const
import grpclib.client

import pymap.admin.grpc.admin_pb2


class AdminBase(abc.ABC):

    @abc.abstractmethod
    async def Append(self, stream):
        pass

    def __mapping__(self):
        return {
            '/admin.Admin/Append': grpclib.const.Handler(
                self.Append,
                grpclib.const.Cardinality.UNARY_UNARY,
                pymap.admin.grpc.admin_pb2.AppendRequest,
                pymap.admin.grpc.admin_pb2.AppendResponse,
            ),
        }


class AdminStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.Append = grpclib.client.UnaryUnaryMethod(
            channel,
            '/admin.Admin/Append',
            pymap.admin.grpc.admin_pb2.AppendRequest,
            pymap.admin.grpc.admin_pb2.AppendResponse,
        )




import unittest

from pymap.parsing.specials.sequenceset import MaxValue, SequenceSet


class TestSequenceSet(unittest.TestCase):

    def test_contains(self) -> None:
        set1 = SequenceSet([12])
        self.assertIn(12, set1.flatten(100))
        self.assertNotIn(13, set1.flatten(100))
        self.assertNotIn(12, set1.flatten(10))
        set2 = SequenceSet([MaxValue()])
        self.assertIn(100, set2.flatten(100))
        self.assertNotIn(21, set2.flatten(100))
        self.assertNotIn(21, set2.flatten(20))
        set3 = SequenceSet([(5, 10)])
        self.assertIn(7, set3.flatten(100))
        self.assertNotIn(7, set3.flatten(6))
        self.assertNotIn(11, set3.flatten(100))
        self.assertNotIn(4, set3.flatten(100))
        set4 = SequenceSet([(MaxValue(), 10)])
        self.assertIn(11, set4.flatten(100))
        self.assertNotIn(9, set4.flatten(100))
        self.assertNotIn(101, set4.flatten(100))
        set5 = SequenceSet([(10, MaxValue())])
        self.assertIn(11, set5.flatten(100))
        self.assertNotIn(9, set5.flatten(100))
        self.assertNotIn(101, set5.flatten(100))
        set6 = SequenceSet([(MaxValue(), MaxValue())])
        self.assertIn(12, set6.flatten(12))
        self.assertNotIn(13, set6.flatten(12))
        self.assertNotIn(11, set6.flatten(12))

    def test_set(self) -> None:
        set1 = SequenceSet([12])
        self.assertEqual([12], list(set1.flatten(100)))
        set2 = SequenceSet([MaxValue()])
        self.assertEqual([100], list(set2.flatten(100)))
        set3 = SequenceSet([(MaxValue(), MaxValue())])
        self.assertEqual([100], list(set3.flatten(100)))
        set4 = SequenceSet([(5, 10)])
        self.assertEqual(list(range(5, 11)), list(set4.flatten(100)))
        set5 = SequenceSet([(MaxValue(), 10)])
        self.assertEqual(list(range(10, 101)), list(set5.flatten(100)))
        set6 = SequenceSet([(10, MaxValue())])
        self.assertEqual(list(range(10, 101)), list(set6.flatten(100)))
        set7 = SequenceSet([(1, 2), (4, 5), (7, 8)])
        self.assertEqual([1, 2, 4, 5, 7, 8], list(set7.flatten(100)))
        set8 = SequenceSet([(3, 5), (4, 6)])
        self.assertEqual([3, 4, 5, 6], list(set8.flatten(100)))
        set9 = SequenceSet([(MaxValue(), 50), (60, 10)])
        self.assertEqual(list(range(10, 101)), list(set9.flatten(100)))
        set10 = SequenceSet([(10, 1000)])
        self.assertEqual(list(range(10, 101)), list(set10.flatten(100)))
        set11 = SequenceSet([(1000, 1000)])
        self.assertEqual([], list(set11.flatten(100)))

    def test_bytes(self) -> None:
        seq = SequenceSet([12, MaxValue(), (1, MaxValue())])
        self.assertEqual(b'12,*,1:*', bytes(seq))
        self.assertEqual(b'12,*,1:*', bytes(seq))

    def test_build(self) -> None:
        seq = SequenceSet.build([1])
        self.assertEqual(b'1', bytes(seq))
        seq = SequenceSet.build([1, 1, 1])
        self.assertEqual(b'1', bytes(seq))
        seq = SequenceSet.build([1, 2, 3, 4, 5])
        self.assertEqual(b'1:5', bytes(seq))
        seq = SequenceSet.build([1, 2, 4, 5])
        self.assertEqual(b'1:2,4:5', bytes(seq))
        seq = SequenceSet.build([1, 2, 4, 6])
        self.assertEqual(b'1:2,4,6', bytes(seq))
        seq = SequenceSet.build([1, 3, 5])
        self.assertEqual(b'1,3,5', bytes(seq))
        seq = SequenceSet.build([1, 2, 3, 4, 5])




import unittest

from pymap.parsing import Params
from pymap.parsing.command.nonauth import AuthenticateCommand, LoginCommand


class TestAuthenticateCommand(unittest.TestCase):

    def test_parse(self):
        ret, buf = AuthenticateCommand.parse(b' PLAIN\n  ', Params())
        self.assertEqual(b'PLAIN', ret.mech_name)
        self.assertEqual(b'  ', buf)


class TestLoginCommand(unittest.TestCase):

    def test_parse(self):
        ret, buf = LoginCommand.parse(b' one two\n  ', Params())
        self.assertEqual(b'one', ret.userid)
        self.assertEqual(b'two', ret.password)
        self.assertEqual(b'  ', buf)




import asyncio
import binascii
import logging
import re
import sys
from argparse import ArgumentParser
from asyncio import shield, Task, StreamReader, StreamWriter, AbstractServer, \
    CancelledError
from base64 import b64encode, b64decode
from ssl import SSLContext
from typing import TypeVar, Iterable, Sequence, List, Optional, Awaitable

from pymap.concurrent import Event
from pymap.config import IMAPConfig
from pymap.context import subsystem, current_command, socket_info
from pymap.exceptions import ResponseError
from pymap.interfaces.backend import BackendInterface, ServiceInterface
from pymap.interfaces.session import LoginProtocol
from pymap.parsing.command import Command
from pymap.parsing.commands import Commands
from pymap.parsing.command.nonauth import AuthenticateCommand, StartTLSCommand
from pymap.parsing.command.select import IdleCommand
from pymap.parsing.exceptions import RequiresContinuation
from pymap.parsing.response import ResponseContinuation, Response, \
    ResponseCode, ResponseBad, ResponseNo, ResponseBye, ResponseOk
from pymap.sockets import InheritedSockets, SocketInfo
from pysasl import ServerChallenge, AuthenticationError, \
    AuthenticationCredentials

from .state import ConnectionState

__all__ = ['IMAPService', 'IMAPServer', 'IMAPConnection']

_Ret = TypeVar('_Ret')
_log = logging.getLogger(__name__)


class IMAPService(ServiceInterface):
    """A pymap service implementing an IMAP server."""

    def __init__(self, servers: Sequence[AbstractServer]) -> None:
        super().__init__()
        self._task = asyncio.create_task(self._run_all(servers))

    @property
    def task(self) -> Task:
        return self._task

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        group = parser.add_argument_group('imap service')
        group.add_argument('--host', metavar='IFACE', action='append',
                           help='the network interface to listen on')
        group.add_argument('--port', metavar='NUM', default='imap',
                           help='the port or service name to listen on')
        group.add_argument('--cert', metavar='FILE', help='cert file for TLS')
        group.add_argument('--key', metavar='FILE', help='key file for TLS')
        group.add_argument('--systemd-sockets', action='store_const',
                           dest='inherited_sockets', const='systemd',
                           help='use systemd inherited sockets')
        group.add_argument('--insecure-login', action='store_true',
                           help='allow plaintext login without TLS')

    @classmethod
    async def start(cls, backend: BackendInterface,
                    config: IMAPConfig) -> 'IMAPService':
        imap_server = IMAPServer(backend.login, config)
        if config.args.inherited_sockets:
            sockets = InheritedSockets.of(config.args.inherited_sockets).get()
            if not sockets:
                raise ValueError('No inherited sockets found')
            return cls([await asyncio.start_server(imap_server, sock=sock)
                        for sock in sockets])
        else:
            server = await asyncio.start_server(
                imap_server, host=config.host, port=config.port)
            return cls([server])

    @classmethod
    async def _run_all(cls, servers: Sequence[AbstractServer]) -> None:
        await asyncio.gather(*[cls._run(server) for server in servers])

    @classmethod
    async def _run(cls, server: AbstractServer) -> None:
        async with server:  # type: ignore
            await server.serve_forever()  # type: ignore


class IMAPServer:
    """Callable object that creates and runs :class:`IMAPConnection` objects
    when :func:`asyncio.start_server` receives a new connection.

    Args:
        login: Login callback that takes authentication credentials and returns
            a :class:`~pymap.interfaces.session.SessionInterface` object.
        config: Settings to use for the IMAP server.

    """

    __slots__ = ['commands', '_login', '_config']

    def __init__(self, login: LoginProtocol, config: IMAPConfig) -> None:
        super().__init__()
        self.commands = config.commands
        self._login = login
        self._config = config

    async def __call__(self, reader: StreamReader,
                       writer: StreamWriter) -> None:
        conn = IMAPConnection(self.commands, self._config, reader, writer)
        state = ConnectionState(self._login, self._config)
        try:
            await conn.run(state)
        finally:
            writer.close()


class IMAPConnection:
    """Runs a single IMAP connection from start to finish.

    Args:
        commands: Defines the IMAP commands available to the connection.
        config: Settings to use for the IMAP connection.
        reader: The input stream for the socket.
        writer: The output stream for the socket.

    """

    _lines = re.compile(br'\r?\n')
    _literal_plus = re.compile(br'{(\d+)\+}\r?\n$')

    __slots__ = ['commands', 'config', 'params', 'bad_command_limit',
                 'reader', 'writer']

    def __init__(self, commands: Commands, config: IMAPConfig,
                 reader: StreamReader,
                 writer: StreamWriter) -> None:
        super().__init__()
        self.commands = commands
        self.config = config
        self.params = config.parsing_params
        self.bad_command_limit = config.bad_command_limit
        self._reset_streams(reader, writer)

    def _reset_streams(self, reader: StreamReader,
                       writer: StreamWriter) -> None:
        self.reader = reader
        self.writer = writer
        socket_info.set(SocketInfo(writer))

    @classmethod
    def _print(cls, log_format: str, output: bytes) -> None:
        if _log.isEnabledFor(logging.DEBUG):
            fd = socket_info.get().socket.fileno()
            lines = cls._lines.split(output)
            if not lines[-1]:
                lines = lines[:-1]
            for line in lines:
                line_str = str(line, 'utf-8', 'replace')
                _log.debug(log_format, fd, line_str)

    def _exec(self, future: Awaitable[_Ret]) -> Awaitable[_Ret]:
        return subsystem.get().execute(future)

    async def readline(self) -> memoryview:
        buf = bytearray(await self.reader.readline())
        while True:
            if not buf.endswith(b'\n'):
                raise EOFError()
            elif buf.endswith(b'+}\n') or buf.endswith(b'+}\r\n'):
                lit_plus = self._literal_plus.search(buf)
            else:
                lit_plus = None
            if lit_plus:
                literal_length = int(lit_plus.group(1))
                buf += await self.reader.readexactly(literal_length)
                buf += await self.reader.readline()
            else:
                self._print('%d -->| %s', buf)
                return memoryview(buf)

    async def read_continuation(self, literal_length: int) -> memoryview:
        extra_literal = await self.reader.readexactly(literal_length)
        self._print('%d -->| %s', extra_literal)
        extra_line = await self.readline()
        extra = extra_literal + bytes(extra_line)
        return memoryview(extra)

    async def authenticate(self, state: ConnectionState, mech_name: bytes) \
            -> Optional[AuthenticationCredentials]:
        mech = state.auth.get_server(mech_name)
        if not mech:
            return None
        responses: List[ServerChallenge] = []
        while True:
            try:
                creds, final = mech.server_attempt(responses)
            except ServerChallenge as chal:
                chal_bytes = b64encode(chal.get_challenge())
                cont = ResponseContinuation(chal_bytes)
                await self.write_response(cont)
                resp_bytes = bytes(await self.read_continuation(0))
                try:
                    chal.set_response(b64decode(resp_bytes))
                except binascii.Error as exc:
                    raise AuthenticationError(exc)
                if resp_bytes.rstrip(b'\r\n') == b'*':
                    raise AuthenticationError('Authentication canceled.')
                responses.append(chal)
            else:
                if final is not None:
                    cont = ResponseContinuation(b64encode(final))
                    await self.write_response(cont)
                    await self.read_continuation(0)
                return creds

    async def read_command(self) -> Command:
        line = await self.readline()
        conts: List[memoryview] = []
        while True:
            try:
                cmd, _ = self.commands.parse(
                    line, self.params.copy(continuations=conts.copy()))
            except RequiresContinuation as req:
                cont = ResponseContinuation(req.message)
                await self.write_response(cont)
                ret = await self.read_continuation(req.literal_length)
                conts.append(ret)
            else:
                return cmd

    async def read_idle_done(self, cmd: IdleCommand) -> bool:
        buf = await self.read_continuation(0)
        ok, _ = cmd.parse_done(buf)
        return ok

    async def write_response(self, resp: Response) -> None:
        resp.write(self.writer)
        try:
            await self.writer.drain()
        except ConnectionError:
            pass
        else:
            self._print('%d <--| %s', bytes(resp))

    async def start_tls(self, ssl_context: SSLContext) -> None:
        loop = asyncio.get_event_loop()
        transport = self.writer.transport
        protocol = transport.get_protocol()  # type: ignore
        new_transport = await loop.start_tls(  # type: ignore
            transport, protocol, ssl_context, server_side=True)
        protocol._stream_reader = StreamReader(loop=loop)
        protocol._client_connected_cb = self._reset_streams
        protocol.connection_made(new_transport)
        self._print('%d <->| %s', b'<TLS handshake>')

    async def send_error_disconnect(self) -> None:
        exc_type, exc, exc_tb = sys.exc_info()
        if isinstance(exc, CancelledError):
            resp = ResponseBye(b'Server has closed the connection.',
                               ResponseCode.of(b'UNAVAILABLE'))
        else:
            resp = ResponseBye(b'Unhandled server error.',
                               ResponseCode.of(b'SERVERBUG'))
        try:
            await self.write_response(resp)
        except IOError:
            pass

    async def write_updates(self, untagged: Iterable[Response]) -> None:
        for i, resp in enumerate(untagged):
            await self.write_response(resp)

    async def handle_updates(self, state: ConnectionState, done: Event,
                             cmd: IdleCommand) -> None:
        timeout = self.config.max_idle_wait
        while not done.is_set():
            receive_task = asyncio.create_task(
                self._exec(state.receive_updates(cmd, done)))
            try:
                untagged = await asyncio.wait_for(receive_task, timeout)
            except TimeoutError:
                pass
            else:
                await shield(self.write_updates(untagged))

    async def idle(self, state: ConnectionState, cmd: IdleCommand) -> Response:
        response = await self._exec(state.do_command(cmd))
        if not isinstance(response, ResponseOk):
            return response
        await self.write_response(ResponseContinuation(b'Idling.'))
        done = subsystem.get().new_event()
        updates_task = asyncio.create_task(
            self.handle_updates(state, done, cmd))
        done_task = asyncio.create_task(self.read_idle_done(cmd))
        updates_exc: Optional[Exception] = None
        done_exc: Optional[Exception] = None
        try:
            ok = await done_task
        except Exception as exc:
            done_exc = exc
        finally:
            done.set()
        try:
            await updates_task
        except Exception as exc:
            updates_exc = exc
        if updates_exc:
            raise updates_exc
        elif done_exc:
            raise done_exc
        elif not ok:
            return ResponseBad(cmd.tag, b'Expected "DONE".')
        else:
            return response

    async def run(self, state: ConnectionState) -> None:
        """Start the socket communication with the IMAP greeting, and then
        enter the command/response cycle.

        Args:
            state: Defines the interaction with the backend plugin.

        """
        self._print('%d +++| %s', bytes(socket_info.get()))
        bad_commands = 0
        try:
            greeting = await self._exec(state.do_greeting())
        except ResponseError as exc:
            resp = exc.get_response(b'*')
            resp.condition = ResponseBye.condition
            await self.write_response(resp)
            return
        else:
            await self.write_response(greeting)
        while True:
            try:
                cmd = await self.read_command()
            except (ConnectionError, EOFError):
                break
            except CancelledError:
                await self.send_error_disconnect()
                break
            except Exception:
                await self.send_error_disconnect()
                raise
            else:
                prev_cmd = current_command.set(cmd)
                try:
                    if isinstance(cmd, AuthenticateCommand):
                        creds = await self.authenticate(state, cmd.mech_name)
                        response, _ = await self._exec(
                            state.do_authenticate(cmd, creds))
                    elif isinstance(cmd, IdleCommand):
                        response = await self.idle(state, cmd)
                    else:
                        response = await self._exec(state.do_command(cmd))
                except ResponseError as exc:
                    resp = exc.get_response(cmd.tag)
                    await self.write_response(resp)
                    if resp.is_terminal:
                        break
                except AuthenticationError as exc:
                    msg = bytes(str(exc), 'utf-8', 'surrogateescape')
                    resp = ResponseBad(cmd.tag, msg)
                    await self.write_response(resp)
                except TimeoutError:
                    resp = ResponseNo(cmd.tag, b'Operation timed out.',
                                      ResponseCode.of(b'TIMEOUT'))
                    await self.write_response(resp)
                except CancelledError:
                    await self.send_error_disconnect()
                    break
                except Exception:
                    await self.send_error_disconnect()
                    raise
                else:
                    await self.write_response(response)
                    if response.is_bad:
                        bad_commands += 1
                        if self.bad_command_limit \
                                and bad_commands >= self.bad_command_limit:
                            msg = b'Too many errors, disconnecting.'
                            response.add_untagged(ResponseBye(msg))
                    else:
                        bad_commands = 0
                    if response.is_terminal:
                        break
                    if isinstance(cmd, StartTLSCommand) and state.ssl_context \
                            and isinstance(response, ResponseOk):
                        await self.start_tls(state.ssl_context)
                finally:
                    await state.do_cleanup()
                    current_command.reset(prev_cmd)
        self._print('%d ---| %s', b'<disconnected>')




import asyncio
import random
from datetime import datetime
from typing import Optional, Sequence, List, Dict, Tuple, FrozenSet, \
    Iterable, Awaitable

from aioredis import Redis, MultiExecError, WatchVariableError  # type: ignore

from pymap.exceptions import MailboxAbort, MailboxNotFound, MailboxConflict
from pymap.flags import FlagOp
from pymap.interfaces.message import AppendMessage, CachedMessage
from pymap.listtree import ListTree
from pymap.mailbox import MailboxSnapshot
from pymap.mime import MessageContent
from pymap.parsing.modutf7 import modutf7_encode, modutf7_decode
from pymap.parsing.specials import FetchRequirement, SequenceSet
from pymap.parsing.specials.flag import Flag, Deleted, Seen
from pymap.selected import SelectedSet, SelectedMailbox

from ..mailbox import Message, MailboxDataInterface, MailboxSetInterface

__all__ = ['Message', 'MailboxData', 'MailboxSet']


async def _delete_keys(redis: Redis, prefixes: Iterable[bytes]) -> None:
    for prefix in prefixes:
        cur = b'0'
        match = prefix + b':*'
        while cur:
            cur, keys = await redis.scan(cur, match=match)
            if keys:
                await redis.unlink(*keys)


async def _check_errors(multi) -> bool:
    # Prevents warning about exception never being retrieved.
    errors = await asyncio.gather(*multi._results, return_exceptions=True)
    return any(not isinstance(exc, WatchVariableError) for exc in errors)


class MailboxData(MailboxDataInterface[Message]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxDataInterface`
    for the redis backend.

    """

    def __init__(self, redis: Redis, guid: bytes, prefix: bytes,
                 uid_validity: int) -> None:
        super().__init__()
        self._redis = redis
        self._guid = guid
        self._prefix = prefix
        self._abort_key = prefix + b':abort'
        self._uid_validity = uid_validity
        self._selected_set = SelectedSet()

    @property
    def guid(self) -> bytes:
        return self._guid

    @property
    def readonly(self) -> bool:
        return False

    @property
    def uid_validity(self) -> int:
        return self._uid_validity

    @property
    def selected_set(self) -> SelectedSet:
        return self._selected_set

    def parse_message(self, append_msg: AppendMessage) -> Message:
        return Message.parse(0, append_msg.message, append_msg.flag_set,
                             append_msg.when, recent=True)

    async def update_selected(self, selected: SelectedMailbox) \
            -> SelectedMailbox:
        last_mod_seq = selected.mod_sequence
        if last_mod_seq is None:
            mod_seq, updated, expunged = await self._get_initial()
        else:
            mod_seq, updated, expunged = await self._get_updated(last_mod_seq)
        selected.mod_sequence = mod_seq
        selected.add_updates(updated, expunged)
        return selected

    async def add(self, append_msg: AppendMessage, recent: bool = False) \
            -> Message:
        redis = self._redis
        prefix = self._prefix
        is_deleted = Deleted in append_msg.flag_set
        is_unseen = Seen not in append_msg.flag_set
        msg_content = MessageContent.parse(append_msg.message)
        msg_flags = [flag.value for flag in append_msg.flag_set]
        msg_time = append_msg.when.isoformat().encode('ascii')
        while True:
            await redis.watch(prefix + b':max-mod', self._abort_key)
            max_uid, max_mod, abort = await redis.mget(
                prefix + b':max-uid', prefix + b':max-mod', self._abort_key)
            MailboxAbort.assertFalse(abort)
            new_uid = int(max_uid or 0) + 1
            new_mod = int(max_mod or 0) + 1
            msg_prefix = prefix + b':msg:%d' % new_uid
            multi = redis.multi_exec()
            multi.set(prefix + b':max-uid', new_uid)
            multi.set(prefix + b':max-mod', new_mod)
            multi.sadd(prefix + b':uids', new_uid)
            multi.zadd(prefix + b':mod-sequence', new_mod, new_uid)
            multi.zadd(prefix + b':sequence', new_uid, new_uid)
            if recent:
                multi.sadd(prefix + b':recent', new_uid)
            if is_deleted:
                multi.sadd(prefix + b':deleted', new_uid)
            if is_unseen:
                multi.zadd(prefix + b':unseen', new_uid, new_uid)
            if msg_flags:
                multi.sadd(msg_prefix + b':flags', *msg_flags)
            multi.set(msg_prefix + b':time', msg_time)
            multi.set(msg_prefix + b':header', bytes(msg_content.header))
            multi.set(msg_prefix + b':body', bytes(msg_content.body))
            try:
                await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break
        return Message(new_uid, append_msg.flag_set, append_msg.when,
                       recent=recent, content=msg_content)

    async def get(self, uid: int, cached_msg: CachedMessage = None,
                  requirement: FetchRequirement = FetchRequirement.METADATA) \
            -> Optional[Message]:
        redis = self._redis
        prefix = self._prefix
        msg_prefix = prefix + b':msg:%d' % uid
        multi = redis.multi_exec()
        multi.sismember(prefix + b':uids', uid)
        multi.smembers(msg_prefix + b':flags')
        multi.get(msg_prefix + b':time')
        multi.sismember(prefix + b':recent', uid)
        if requirement & FetchRequirement.BODY:
            multi.get(msg_prefix + b':header')
            multi.get(msg_prefix + b':body')
        elif requirement & FetchRequirement.HEADERS:
            multi.get(msg_prefix + b':header')
            multi.echo(b'')
        else:
            multi.echo(b'')
            multi.echo(b'')
        multi.get(self._abort_key)
        exists, flags, time, recent, header, body, abort = \
            await multi.execute()
        MailboxAbort.assertFalse(abort)
        if not exists:
            if cached_msg is None:
                return None
            else:
                return Message(cached_msg.uid, cached_msg.permanent_flags,
                               cached_msg.internal_date, expunged=True)
        msg_flags = {Flag(flag) for flag in flags}
        msg_time = datetime.fromisoformat(time.decode('ascii'))
        msg_recent = bool(recent)
        if header:
            msg_content = MessageContent.parse_split(header, body)
            return Message(uid, msg_flags, msg_time, recent=msg_recent,
                           content=msg_content)
        else:
            return Message(uid, msg_flags, msg_time, recent=msg_recent)

    async def delete(self, uids: Iterable[int]) -> None:
        redis = self._redis
        prefix = self._prefix
        uids = list(uids)
        if not uids:
            return
        while True:
            await redis.watch(prefix + b':max-mod', self._abort_key)
            max_mod, abort = await redis.mget(prefix + b':max-mod',
                                              self._abort_key)
            MailboxAbort.assertFalse(abort)
            new_mod = int(max_mod or 0) + 1
            multi = redis.multi_exec()
            multi.set(prefix + b':max-mod', new_mod)
            for uid in uids:
                multi.zadd(prefix + b':expunged', new_mod, uid)
            multi.srem(prefix + b':uids', *uids)
            multi.zrem(prefix + b':sequence', *uids)
            multi.zrem(prefix + b':mod-sequence', *uids)
            multi.srem(prefix + b':recent', *uids)
            multi.srem(prefix + b':deleted', *uids)
            multi.zrem(prefix + b':unseen', *uids)
            multi.sadd(prefix + b':cleanup', *uids)
            try:
                await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break

    async def claim_recent(self, selected: SelectedMailbox) -> None:
        redis = self._redis
        prefix = self._prefix
        while True:
            await redis.watch(prefix + b':max-mod', self._abort_key)
            recent = await redis.smembers(prefix + b':recent')
            if not recent:
                break
            max_mod, abort = await redis.mget(prefix + b':max-mod',
                                              self._abort_key)
            MailboxAbort.assertFalse(abort)
            new_mod = int(max_mod or 0) + 1
            multi = self._redis.multi_exec()
            multi.set(prefix + b':max-mod', new_mod)
            for uid in recent:
                multi.zadd(prefix + b':mod-sequence', new_mod, uid)
            multi.delete(prefix + b':recent')
            try:
                await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break
        for uid_bytes in recent:
            selected.session_flags.add_recent(int(uid_bytes))

    async def update_flags(self, messages: Sequence[Message],
                           flag_set: FrozenSet[Flag], mode: FlagOp) -> None:
        redis = self._redis
        prefix = self._prefix
        messages = list(messages)
        if not messages:
            return
        uids = {msg.uid: msg for msg in messages}
        while True:
            await redis.watch(prefix + b':max-mod', self._abort_key)
            pipe = redis.pipeline()
            pipe.smembers(prefix + b':uids')
            pipe.get(prefix + b':max-mod')
            pipe.get(self._abort_key)
            existing_uids, max_mod, abort = await pipe.execute()
            MailboxAbort.assertFalse(abort)
            update_uids = uids.keys() & {int(uid) for uid in existing_uids}
            if not update_uids:
                break
            new_mod = int(max_mod or 0) + 1
            new_flags: Dict[int, Awaitable[Sequence[bytes]]] = {}
            multi = redis.multi_exec()
            multi.set(prefix + b':max-mod', new_mod)
            for msg in messages:
                msg_uid = msg.uid
                if msg_uid not in update_uids:
                    continue
                msg_prefix = prefix + b':msg:%d' % msg_uid
                flag_vals = (flag.value for flag in flag_set)
                multi.zadd(prefix + b':mod-sequence', new_mod, msg_uid)
                if mode == FlagOp.REPLACE:
                    multi.unlink(msg_prefix + b':flags')
                    if flag_set:
                        multi.sadd(msg_prefix + b':flags', *flag_vals)
                    if Deleted in flag_set:
                        multi.sadd(prefix + b':deleted', msg_uid)
                    else:
                        multi.srem(prefix + b':deleted', msg_uid)
                    if Seen not in flag_set:
                        multi.zadd(prefix + b':unseen', msg_uid, msg_uid)
                    else:
                        multi.zrem(prefix + b':unseen', msg_uid)
                elif mode == FlagOp.ADD and flag_set:
                    multi.sadd(msg_prefix + b':flags', *flag_vals)
                    if Deleted in flag_set:
                        multi.sadd(prefix + b':deleted', msg_uid)
                    if Seen in flag_set:
                        multi.zrem(prefix + b':unseen', msg_uid)
                elif mode == FlagOp.DELETE and flag_set:
                    multi.srem(msg_prefix + b':flags', *flag_vals)
                    if Deleted in flag_set:
                        multi.srem(prefix + b':deleted', msg_uid)
                    if Seen in flag_set:
                        multi.zadd(prefix + b':unseen', msg_uid, msg_uid)
                new_flags[msg_uid] = multi.smembers(msg_prefix + b':flags')
            try:
                await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                for msg_uid, msg_flags in new_flags.items():
                    msg = uids[msg_uid]
                    msg.permanent_flags = frozenset(
                        Flag(flag) for flag in await msg_flags)
                break

    async def cleanup(self) -> None:
        redis = self._redis
        prefix = self._prefix
        while True:
            # TODO: use spop with count= when aioredis > 1.2.0 is tagged
            uids = await redis.srandmember(prefix + b':cleanup', 100)
            if not uids:
                break
            prefixes = (prefix + b':msg:%b' % uid for uid in uids)
            await _delete_keys(redis, prefixes)
            await redis.srem(prefix + b':cleanup', *uids)

    async def find_deleted(self, seq_set: SequenceSet,
                           selected: SelectedMailbox) -> Sequence[int]:
        if not seq_set.is_all:
            return await super().find_deleted(seq_set, selected)
        redis = self._redis
        prefix = self._prefix
        deleted = await redis.smembers(prefix + b':deleted')
        return [int(uid) for uid in deleted]

    async def snapshot(self) -> MailboxSnapshot:
        redis = self._redis
        prefix = self._prefix
        while True:
            await redis.watch(prefix + b':sequence', self._abort_key)
            pipe = redis.pipeline()
            pipe.get(prefix + b':max-uid')
            pipe.zcard(prefix + b':sequence')
            pipe.scard(prefix + b':recent')
            pipe.zcard(prefix + b':unseen')
            pipe.zrange(prefix + b':unseen', 0, 0)
            pipe.get(self._abort_key)
            max_uid, exists, num_recent, num_unseen, unseen, abort = \
                await pipe.execute()
            MailboxAbort.assertFalse(abort)
            next_uid = int(max_uid or 0) + 1
            if not unseen:
                first_unseen: Optional[int] = None
                break
            else:
                first_uid = int(unseen[0])
                multi = redis.multi_exec()
                multi.zrank(prefix + b':sequence', first_uid)
                try:
                    [first_unseen] = await multi.execute()
                except MultiExecError:
                    if await _check_errors(multi):
                        raise
                else:
                    break
        return MailboxSnapshot(self.guid, self.readonly, self.uid_validity,
                               self.permanent_flags, self.session_flags,
                               exists, num_recent, num_unseen, first_unseen,
                               next_uid)

    async def _get_initial(self) \
            -> Tuple[int, Sequence[Message], Sequence[int]]:
        redis = self._redis
        prefix = self._prefix
        while True:
            await redis.watch(prefix + b':max-mod', self._abort_key)
            pipe = redis.pipeline()
            pipe.zrange(prefix + b':sequence')
            pipe.get(self._abort_key)
            uids, abort = await pipe.execute()
            MailboxAbort.assertFalse(abort)
            multi = redis.multi_exec()
            multi.get(prefix + b':max-mod')
            for uid in uids:
                msg_prefix = prefix + b':msg:' + uid
                multi.echo(uid)
                multi.smembers(msg_prefix + b':flags')
                multi.get(msg_prefix + b':time')
            try:
                results = await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break
        mod_seq = int(results[0] or 0)
        updated: List[Message] = []
        for i in range(1, len(results), 3):
            msg_uid = int(results[i])
            msg_flags = {Flag(flag) for flag in results[i + 1]}
            msg_time = datetime.fromisoformat(results[i + 2].decode('ascii'))
            msg = Message(msg_uid, msg_flags, msg_time)
            updated.append(msg)
        return mod_seq, updated, []

    async def _get_updated(self, last_mod_seq: int) \
            -> Tuple[int, Sequence[Message], Sequence[int]]:
        redis = self._redis
        prefix = self._prefix
        while True:
            await redis.watch(prefix + b':max-mod', self._abort_key)
            pipe = redis.pipeline()
            pipe.zrangebyscore(prefix + b':mod-sequence', last_mod_seq)
            pipe.get(self._abort_key)
            uids, abort = await pipe.execute()
            MailboxAbort.assertFalse(abort)
            multi = redis.multi_exec()
            multi.get(prefix + b':max-mod')
            multi.zrangebyscore(prefix + b':expunged', last_mod_seq)
            for uid in uids:
                msg_prefix = prefix + b':msg:' + uid
                multi.echo(uid)
                multi.smembers(msg_prefix + b':flags')
                multi.get(msg_prefix + b':time')
            try:
                results = await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break
        mod_seq = int(results[0] or 0)
        expunged = [int(uid) for uid in results[1]]
        updated: List[Message] = []
        for i in range(2, len(results), 3):
            msg_uid = int(results[i])
            msg_flags = {Flag(flag) for flag in results[i + 1]}
            msg_time = datetime.fromisoformat(results[i + 2].decode('ascii'))
            msg = Message(msg_uid, msg_flags, msg_time)
            updated.append(msg)
        return mod_seq, updated, expunged


class MailboxSet(MailboxSetInterface[MailboxData]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxSetInterface`
    for the redis backend.

    """

    def __init__(self, redis: Redis, prefix: bytes) -> None:
        super().__init__()
        self._redis = redis
        self._prefix = prefix
        self._mbx_key = prefix + b':mailboxes'
        self._max_order_key = prefix + b':mbx-max-order'
        self._order_key = prefix + b':mbx-order'
        self._uidv_key = prefix + b':uid-validity'
        self._sub_key = prefix + b':subscribed'

    @property
    def delimiter(self) -> str:
        return '/'

    async def set_subscribed(self, name: str, subscribed: bool) -> None:
        if subscribed:
            await self._redis.sadd(self._sub_key, name)
        else:
            await self._redis.srem(self._sub_key, name)

    async def list_subscribed(self) -> ListTree:
        mailboxes = [modutf7_decode(name) for name in
                     await self._redis.smembers(self._sub_key)]
        return ListTree(self.delimiter).update('INBOX', *mailboxes)

    async def list_mailboxes(self) -> ListTree:
        multi = self._redis.multi_exec()
        multi.hgetall(self._mbx_key)
        multi.zrange(self._order_key)
        mailboxes, mbx_order = await multi.execute()
        rev_mbx = {guid: key for key, guid in mailboxes.items()}
        mailboxes = [modutf7_decode(rev_mbx[guid]) for guid in mbx_order
                     if guid in rev_mbx]
        return ListTree(self.delimiter).update('INBOX', *mailboxes)

    async def get_mailbox(self, name: str,
                          try_create: bool = False) -> 'MailboxData':
        redis = self._redis
        name_key = modutf7_encode(name)
        while True:
            await redis.watch(self._mbx_key)
            mbx_guid = await redis.hget(self._mbx_key, name_key)
            if mbx_guid is None:
                raise MailboxNotFound(name, try_create)
            multi = redis.multi_exec()
            multi.hget(self._uidv_key, name_key)
            try:
                uidval, *_ = await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                mbx_prefix = b'%b:%b' % (self._prefix, mbx_guid)
                return MailboxData(redis, mbx_guid, mbx_prefix, int(uidval))

    async def add_mailbox(self, name: str) -> None:
        redis = self._redis
        name_key = modutf7_encode(name)
        while True:
            await redis.watch(self._mbx_key)
            mbx_guid = b'%032x' % random.getrandbits(128)
            pipe = redis.pipeline()
            pipe.incr(self._max_order_key)
            pipe.hexists(self._mbx_key, name_key)
            order, exists = await pipe.execute()
            if exists:
                raise MailboxConflict(name)
            multi = redis.multi_exec()
            multi.hset(self._mbx_key, name_key, mbx_guid)
            multi.zadd(self._order_key, order, mbx_guid)
            multi.hincrby(self._uidv_key, name_key)
            try:
                _, _, uidval = await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break

    async def delete_mailbox(self, name: str) -> None:
        redis = self._redis
        name_key = modutf7_encode(name)
        multi = redis.multi_exec()
        multi.hget(self._mbx_key, name_key)
        multi.hdel(self._mbx_key, name_key)
        mbx_guid, _ = await multi.execute()
        if mbx_guid is None:
            raise MailboxNotFound(name)
        await redis.zrem(self._order_key, mbx_guid)
        mbx_prefix = b'%b:%b' % (self._prefix, mbx_guid)
        await redis.set(mbx_prefix + b':abort', 1)
        asyncio.create_task(_delete_keys(redis, [mbx_prefix]))

    async def rename_mailbox(self, before: str, after: str) -> None:
        redis = self._redis
        while True:
            await redis.watch(self._mbx_key)
            all_keys = await redis.hgetall(self._mbx_key)
            all_mbx = {modutf7_decode(key): ns for key, ns in all_keys.items()}
            tree = ListTree(self.delimiter).update('INBOX', *all_mbx.keys())
            before_entry = tree.get(before)
            after_entry = tree.get(after)
            if before_entry is None:
                raise MailboxNotFound(before)
            elif after_entry is not None:
                raise MailboxConflict(after)
            multi = redis.multi_exec()
            for before_name, after_name in tree.get_renames(before, after):
                before_guid = all_mbx[before_name]
                before_key = modutf7_encode(before_name)
                after_key = modutf7_encode(after_name)
                multi.hset(self._mbx_key, after_key, before_guid)
                multi.hdel(self._mbx_key, before_key)
                multi.hincrby(self._uidv_key, after_key)
            if before == 'INBOX':
                inbox_guid = b'%032x' % random.getrandbits(128)
                multi.hset(self._mbx_key, b'INBOX', inbox_guid)
                multi.hincrby(self._uidv_key, b'INBOX')
            try:
                await multi.execute()
            except MultiExecError:
                if await _check_errors(multi):
                    raise
            else:
                break




import unittest

from pymap.mime import MessageContent
from pymap.mime.cte import MessageDecoder

_7bit_body = b"""Testing 7bit\n"""
_8bit_body = b"""Testing\x008bit\x00\n"""
_qp_body = b"""Testing=01Quoted=3DPrintable\n"""
_b64_body = b"""VGVzdGluZwEACkJhc2UgNjQgCg==\n"""


class TestMessageDecoder(unittest.TestCase):

    def test_7bit_cte(self) -> None:
        data = b'\n' + _7bit_body
        msg = MessageContent.parse(data)
        decoded = MessageDecoder.of(msg.header).decode(msg.body)
        self.assertEqual(b'Testing 7bit\n', bytes(decoded))

    def test_8bit_cte(self) -> None:
        data = b'Content-Transfer-Encoding: 8bit\n\n' + _8bit_body
        msg = MessageContent.parse(data)
        decoded = MessageDecoder.of(msg.header).decode(msg.body)
        self.assertEqual(b'Testing\x008bit\x00\n', bytes(decoded))

    def test_quopri_cte(self) -> None:
        data = b'Content-Transfer-Encoding: quoted-printable\n\n' + _qp_body
        msg = MessageContent.parse(data)
        decoded = MessageDecoder.of(msg.header).decode(msg.body)
        self.assertEqual(b'Testing\x01Quoted=Printable\n', bytes(decoded))

    def test_base64_cte(self) -> None:
        data = b'Content-Transfer-Encoding: base64\n\n' + _b64_body
        msg = MessageContent.parse(data)
        decoded = MessageDecoder.of(msg.header).decode(msg.body)
        self.assertEqual(b'Testing\x01\x00\nBase 64 \n', bytes(decoded))




from abc import abstractmethod
from typing import TypeVar, Optional, Tuple, Sequence, FrozenSet, \
    Iterable, AsyncIterable
from typing_extensions import Protocol

from pymap.flags import FlagOp
from pymap.interfaces.message import AppendMessage, CachedMessage
from pymap.listtree import ListTree
from pymap.mailbox import MailboxSnapshot
from pymap.message import BaseMessage
from pymap.parsing.specials import SequenceSet, FetchRequirement
from pymap.parsing.specials.flag import get_system_flags, Flag, Deleted, Recent
from pymap.selected import SelectedSet, SelectedMailbox

__all__ = ['MailboxDataInterface', 'MailboxSetInterface', 'Message',
           'MessageT', 'MailboxDataT', 'MailboxDataT_co']

#: Type variable with an upper bound of :class:`Message`.
MessageT = TypeVar('MessageT', bound='Message')

#: Type variable with an upper bound of :class:`MailboxDataInterface`.
MailboxDataT = TypeVar('MailboxDataT', bound='MailboxDataInterface')

#: Covariant type variable with an upper bound of
#: :class:`MailboxDataInterface`.
MailboxDataT_co = TypeVar('MailboxDataT_co', bound='MailboxDataInterface',
                          covariant=True)


class Message(BaseMessage):
    """Manages a single message. This message does not have its contents
    (headers, body, etc.) loaded into memory, it is only the IMAP metadata
    such as UID and flags.

    """

    @property
    def recent(self) -> bool:
        """True if the message is considered new in the mailbox. The next
        session to SELECT the mailbox will negate this value and apply the
        ``\\Recent`` session flag to the message.

        """
        return self._kwargs.get('recent', False)

    @recent.setter
    def recent(self, recent: bool) -> None:
        self._kwargs['recent'] = recent


class MailboxDataInterface(Protocol[MessageT]):
    """Manages the messages and metadata associated with a single mailbox."""

    @property
    @abstractmethod
    def guid(self) -> bytes:
        """The mailbox GUID.

        See Also:
            :attr:`~pymap.interfaces.mailbox.MailboxInterface.guid`

        """
        ...

    @property
    @abstractmethod
    def readonly(self) -> bool:
        """Whether the mailbox is read-only or read-write."""
        ...

    @property
    @abstractmethod
    def uid_validity(self) -> int:
        """The mailbox UID validity value."""
        ...

    @property
    def permanent_flags(self) -> FrozenSet[Flag]:
        """The permanent flags allowed in the mailbox."""
        return get_system_flags() - {Recent}

    @property
    def session_flags(self) -> FrozenSet[Flag]:
        """The session flags allowed in the mailbox."""
        return frozenset({Recent})

    @property
    @abstractmethod
    def selected_set(self) -> SelectedSet:
        """The set of selected mailbox sessions currently active."""
        ...

    @abstractmethod
    async def update_selected(self, selected: SelectedMailbox) \
            -> SelectedMailbox:
        """Populates and returns the selected mailbox object with the state
        needed to discover updates.

        Args:
            selected: the selected mailbox object.

        """
        ...

    @abstractmethod
    async def add(self, message: AppendMessage, recent: bool = False) \
            -> MessageT:
        """Adds a new message to the end of the mailbox, returning a copy of
        message with its assigned UID.

        Args:
            message: The new message data.
            recent: True if the message should be marked recent.

        """
        ...

    @abstractmethod
    async def get(self, uid: int, cached_msg: CachedMessage = None,
                  requirement: FetchRequirement = FetchRequirement.METADATA) \
            -> Optional[MessageT]:
        """Return the message with the given UID.

        Args:
            uid: The message UID.
            cached_msg: The last known cached message.
            requirement: The data required from each message.

        Raises:
            IndexError: The UID is not valid in the mailbox.

        """
        ...

    @abstractmethod
    async def delete(self, uids: Iterable[int]) -> None:
        """Delete messages with the given UIDs.

        Args:
            uids: The message UIDs.

        """
        ...

    @abstractmethod
    async def claim_recent(self, selected: SelectedMailbox) -> None:
        """Messages that are newly added to the mailbox are assigned the
        ``\\Recent`` flag in the current selected mailbox session.

        Args:
            selected: The selected mailbox session.

        """
        ...

    @abstractmethod
    async def update_flags(self, messages: Sequence[MessageT],
                           flag_set: FrozenSet[Flag], mode: FlagOp) -> None:
        """Update the permanent flags of each messages.

        Args:
            messages: The message objects.
            flag_set: The set of flags for the update operation.
            flag_op: The mode to change the flags.

        """
        ...

    @abstractmethod
    async def cleanup(self) -> None:
        """Perform any necessary "housekeeping" steps. This may be a slow
        operation, and may run things like garbage collection on the backend.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.check_mailbox`

        """
        ...

    @abstractmethod
    async def snapshot(self) -> MailboxSnapshot:
        """Returns a snapshot of the current state of the mailbox."""
        ...

    async def find(self, seq_set: SequenceSet, selected: SelectedMailbox,
                   requirement: FetchRequirement = FetchRequirement.METADATA) \
            -> AsyncIterable[Tuple[int, MessageT]]:
        """Find the active message UID and message pairs in the mailbox that
        are contained in the given sequences set. Message sequence numbers
        are resolved by the selected mailbox session.

        Args:
            seq_set: The sequence set of the desired messages.
            selected: The selected mailbox session.
            requirement: The data required from each message.

        """
        for seq, cached_msg in selected.messages.get_all(seq_set):
            msg = await self.get(cached_msg.uid, cached_msg, requirement)
            if msg is not None:
                yield (seq, msg)

    async def find_deleted(self, seq_set: SequenceSet,
                           selected: SelectedMailbox) -> Sequence[int]:
        """Return all the active message UIDs that have the ``\\Deleted`` flag.

        Args:
            seq_set: The sequence set of the possible messages.
            selected: The selected mailbox session.

        """
        session_flags = selected.session_flags
        return [msg.uid async for _, msg in self.find(seq_set, selected)
                if Deleted in msg.get_flags(session_flags)]


class MailboxSetInterface(Protocol[MailboxDataT_co]):
    """Manages the set of mailboxes available to the authenticated user."""

    @property
    @abstractmethod
    def delimiter(self) -> str:
        """The delimiter used in mailbox names to indicate hierarchy."""
        ...

    @abstractmethod
    async def set_subscribed(self, name: str, subscribed: bool) -> None:
        """Add or remove the subscribed status of a mailbox.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.subscribe`
            :meth:`~pymap.interfaces.session.SessionInterface.unsubscribe`

        Args:
            name: The name of the mailbox.
            subscribed: True if the mailbox should be subscribed.

        """
        ...

    @abstractmethod
    async def list_subscribed(self) -> ListTree:
        """Return a list of all subscribed mailboxes.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.list_mailboxes`

        """
        ...

    @abstractmethod
    async def list_mailboxes(self) -> ListTree:
        """Return a list of all mailboxes.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.list_mailboxes`

        """
        ...

    @abstractmethod
    async def get_mailbox(self, name: str, try_create: bool = False) \
            -> MailboxDataT_co:
        """Return an existing mailbox by name.

        Args:
            name: The name of the mailbox.
            try_create: True if the operation might succeed if the mailbox
                is created first.

        Raises:
            :exc:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def add_mailbox(self, name: str) -> None:
        """Create a new mailbox.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.create_mailbox`

        Args:
            name: The name of the mailbox.

        Raises:
            :exc:`~pymap.exceptions.MailboxConflict`

        """
        ...

    @abstractmethod
    async def delete_mailbox(self, name: str) -> None:
        """Delete an existing mailbox.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.delete_mailbox`

        Args:
            name: The name of the mailbox.

        Raises:
            :exc:`~pymap.exceptions.MailboxNotFound`
            :exc:`~pymap.exceptions.MailboxHasChildren`

        """
        ...

    @abstractmethod
    async def rename_mailbox(self, before: str, after: str) -> None:
        """Rename an existing mailbox.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.rename_mailbox`

        Args:
            before: The name of the existing mailbox.
            after: The name of the destination mailbox.

        Raises:
            :exc:`~pymap.exceptions.MailboxNotFound`
            :exc:`~pymap.exceptions.MailboxConflict`

        """
        ...




from abc import abstractmethod
from typing import Any, Tuple, Optional, FrozenSet, Iterable, Sequence
from typing_extensions import Protocol

from pysasl import AuthenticationCredentials

from .filter import FilterSetInterface
from .message import AppendMessage, MessageInterface
from .mailbox import MailboxInterface
from ..concurrent import Event
from ..config import ConfigT_contra
from ..flags import FlagOp
from ..parsing.specials import SequenceSet, FetchAttribute, Flag, SearchKey
from ..parsing.response.code import AppendUid, CopyUid
from ..selected import SelectedMailbox

__all__ = ['LoginProtocol', 'SessionInterface']


class LoginProtocol(Protocol[ConfigT_contra]):
    """Defines the callback protocol that backends use to initialize a new
    session.

    """

    async def __call__(self, credentials: AuthenticationCredentials,
                       config: ConfigT_contra) -> 'SessionInterface':
        """Given a set of authentication credentials, initialize and return a
        new IMAP session for the user.

        Args:
            credentials: Authentication credentials supplied by the user.
            config: The config in use by the server.

        Raises:
            :class:`~pymap.exceptions.InvalidAuth`

        """
        ...


class SessionInterface(Protocol):
    """Corresponds to a single, authenticated IMAP session."""

    @property
    @abstractmethod
    def owner(self) -> str:
        """The SASL authorization identity of the logged-in user."""
        ...

    @property
    @abstractmethod
    def filter_set(self) -> Optional[FilterSetInterface[Any]]:
        """Manages the active and inactive filters for the login user."""
        ...

    @abstractmethod
    async def cleanup(self) -> None:
        """Called after ever operation, success or failure, to allow the
        backend to clean up resources.

        Note:
            Any exception raised by this method will be silently ignored.

        """
        ...

    @abstractmethod
    async def list_mailboxes(self, ref_name: str, filter_: str,
                             subscribed: bool = False,
                             selected: SelectedMailbox = None) \
            -> Tuple[Iterable[Tuple[str, Optional[str], Sequence[bytes]]],
                     Optional[SelectedMailbox]]:
        """List the mailboxes owned by the user.

        See Also:
            `RFC 3501 6.3.8.
            <https://tools.ietf.org/html/rfc3501#section-6.3.8>`_,
            `RFC 3501 6.3.9.
            <https://tools.ietf.org/html/rfc3501#section-6.3.9>`_

        Args:
            ref_name: Mailbox reference name.
            filter_: Mailbox name with possible wildcards.
            subscribed: If True, only list the subscribed mailboxes.
            selected: If applicable, the currently selected mailbox name.

        """
        ...

    @abstractmethod
    async def get_mailbox(self, name: str, selected: SelectedMailbox = None) \
            -> Tuple[MailboxInterface, Optional[SelectedMailbox]]:
        """Retrieves a :class:`~pymap.interfaces.mailbox.MailboxInterface`
        object corresponding to an existing mailbox owned by the user. Raises
        an exception if the mailbox does not yet exist.

        Args:
            name: The name of the mailbox.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def create_mailbox(self, name: str,
                             selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        """Creates a new mailbox owned by the user.

        See Also:
            `RFC 3501 6.3.3.
            <https://tools.ietf.org/html/rfc3501#section-6.3.3>`_

        Args:
            name: The name of the mailbox.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxConflict`

        """
        ...

    @abstractmethod
    async def delete_mailbox(self, name: str,
                             selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        """Deletes the mailbox owned by the user.

        See Also:
            `RFC 3501 6.3.4.
            <https://tools.ietf.org/html/rfc3501#section-6.3.4>`_

        Args:
            name: The name of the mailbox.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.MailboxHasChildren`

        """
        ...

    @abstractmethod
    async def rename_mailbox(self, before_name: str, after_name: str,
                             selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        """Renames the mailbox owned by the user.

        See Also:
            `RFC 3501 6.3.5.
            <https://tools.ietf.org/html/rfc3501#section-6.3.5>`_

        Args:
            before_name: The name of the mailbox before the rename.
            after_name: The name of the mailbox after the rename.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.MailboxConflict`

        """
        ...

    @abstractmethod
    async def subscribe(self, name: str, selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        """Mark the given folder name as subscribed, whether or not the given
        folder name currently exists.

        See Also:
            `RFC 3501 6.3.6.
            <https://tools.ietf.org/html/rfc3501#section-6.3.6>`_

        Args:
            name: The name of the mailbox.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def unsubscribe(self, name: str, selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        """Remove the given folder name from the subscription list, whether or
        not the given folder name currently exists.

        See Also:
            `RFC 3501 6.3.6.
            <https://tools.ietf.org/html/rfc3501#section-6.3.6>`_

        Args:
            name: The name of the mailbox.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def append_messages(self, name: str,
                              messages: Sequence[AppendMessage],
                              selected: SelectedMailbox = None) \
            -> Tuple[AppendUid, Optional[SelectedMailbox]]:
        """Appends a message to the end of the mailbox.

        See Also:
            `RFC 3502 6.3.11.
            <https://tools.ietf.org/html/rfc3502#section-6.3.11>`_

        Args:
            name: The name of the mailbox.
            messages: The messages to append.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.AppendFailure`

        """
        ...

    @abstractmethod
    async def select_mailbox(self, name: str, readonly: bool = False) \
            -> Tuple[MailboxInterface, SelectedMailbox]:
        """Selects an existing mailbox owned by the user. The returned session
        is then used as the ``selected`` argument to other methods to fetch
        mailbox updates.

        See Also:
            `RFC 3501 6.3.1.
            <https://tools.ietf.org/html/rfc3501#section-6.3.1>`_,
            `RFC 3501 6.3.2.
            <https://tools.ietf.org/html/rfc3501#section-6.3.2>`_

        Args:
            name: The name of the mailbox.
            readonly: True if the mailbox is read-only.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def check_mailbox(self, selected: SelectedMailbox, *,
                            wait_on: Event = None,
                            housekeeping: bool = False) -> SelectedMailbox:
        """Checks for any updates in the mailbox.

        If ``wait_on`` is given, this method should block until either this
        event is signalled or the mailbox has detected updates. This method may
        be called continuously as long as ``wait_on`` is not signalled.

        If ``housekeeping`` is True, perform any house-keeping necessary by the
        mailbox backend, which may be a slower operation.

        See Also:
            `RFC 3501 6.1.2.
            <https://tools.ietf.org/html/rfc3501#section-6.1.2>`_,
            `RFC 3501 6.4.1.
            <https://tools.ietf.org/html/rfc3501#section-6.4.1>`_
            `RFC 2177 <https://tools.ietf.org/html/rfc2177>`_

        Args:
            selected: The selected mailbox session.
            wait_on: If given, block until this event signals.
            housekeeping: If True, the backend may perform additional
                housekeeping operations if necessary.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def fetch_messages(self, selected: SelectedMailbox,
                             sequence_set: SequenceSet,
                             attributes: FrozenSet[FetchAttribute]) \
            -> Tuple[Iterable[Tuple[int, MessageInterface]], SelectedMailbox]:
        """Get a list of loaded message objects corresponding to given sequence
        set.

        Args:
            selected: The selected mailbox session.
            sequence_set: Sequence set of message sequences or UIDs.
            attributes: Fetch attributes for the messages.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def search_mailbox(self, selected: SelectedMailbox,
                             keys: FrozenSet[SearchKey]) \
            -> Tuple[Iterable[Tuple[int, MessageInterface]], SelectedMailbox]:
        """Get the messages in the current mailbox that meet all of the
        given search criteria.

        See Also:
            `RFC 3501 7.2.5.
            <https://tools.ietf.org/html/rfc3501#section-7.2.5>`_

        Args:
            selected: The selected mailbox session.
            keys: Search keys specifying the message criteria.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def expunge_mailbox(self, selected: SelectedMailbox,
                              uid_set: SequenceSet = None) -> SelectedMailbox:
        """All messages that are marked as deleted are immediately expunged
        from the mailbox.

        See Also:
            `RFC 3501 6.4.3.
            <https://tools.ietf.org/html/rfc3501#section-6.4.3>`_
            `RFC 4315 2.1 <https://tools.ietf.org/html/rfc4315#section-2.1>`_

        Args:
            selected: The selected mailbox session.
            uid_set: Only the messages in the given UID set should be expunged.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.MailboxReadOnly`

        """
        ...

    @abstractmethod
    async def copy_messages(self, selected: SelectedMailbox,
                            sequence_set: SequenceSet,
                            mailbox: str) \
            -> Tuple[Optional[CopyUid], SelectedMailbox]:
        """Copy a set of messages into the given mailbox.

        See Also:
            `RFC 3501 6.4.7.
            <https://tools.ietf.org/html/rfc3501#section-6.4.7>`_

        Args:
            selected: The selected mailbox session.
            sequence_set: Sequence set of message sequences or UIDs.
            mailbox: Name of the mailbox to copy messages into.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.MailboxReadOnly`

        """
        ...

    @abstractmethod
    async def update_flags(self, selected: SelectedMailbox,
                           sequence_set: SequenceSet,
                           flag_set: FrozenSet[Flag],
                           mode: FlagOp = FlagOp.REPLACE) \
            -> Tuple[Iterable[Tuple[int, MessageInterface]], SelectedMailbox]:
        """Update the flags for the given set of messages.

        See Also:
            `RFC 3501 6.4.6.
            <https://tools.ietf.org/html/rfc3501#section-6.4.6>`_

        Args:
            selected: The selected mailbox session.
            sequence_set: Sequence set of message sequences or UIDs.
            flag_set: Set of flags to update.
            mode: Update mode for the flag set.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.MailboxReadOnly`

        """
        ...



"""Module containing the general exceptions that may be used by pymap."""

from abc import abstractmethod, ABCMeta
from typing import Any, Optional

from .parsing.response import Response, ResponseCode, ResponseNo, ResponseOk, \
    ResponseBye

__all__ = ['ResponseError', 'NotSupportedError', 'CloseConnection',
           'CommandNotAllowed', 'SearchNotAllowed', 'InvalidAuth',
           'MailboxError', 'MailboxAbort', 'MailboxNotFound',
           'MailboxConflict', 'MailboxHasChildren', 'MailboxReadOnly',
           'AppendFailure']


class ResponseError(Exception, metaclass=ABCMeta):
    """The base exception for all custom errors that are used to generate a
    response to an IMAP command.

    """

    @abstractmethod
    def get_response(self, tag: bytes) -> Response:
        """Build an IMAP response for the error.

        Args:
            tag: The command tag that generated the error.

        """
        ...


class NotSupportedError(ResponseError, NotImplementedError):
    """Raised when an action is taken that is not supported by the server or
    backend.

    """

    def get_response(self, tag: bytes) -> ResponseNo:
        return ResponseNo(tag, b'Operation not supported.',
                          ResponseCode.of(b'CANNOT'))


class CloseConnection(ResponseError):
    """Raised when the connection should be closed immediately after sending
    the provided response.

    """

    def get_response(self, tag: bytes) -> ResponseOk:
        response = ResponseOk(tag, b'Logout successful.')
        response.add_untagged(ResponseBye(b'Logging out.'))
        return response


class CommandNotAllowed(ResponseError):
    """The command is syntactically valid, but not allowed.

    Args:
        message: The message to display in the response.
        code: Optional response code for the error.

    """

    def __init__(self, message: bytes, code: ResponseCode = None) -> None:
        super().__init__()
        self.message = message
        self.code = code

    def get_response(self, tag: bytes) -> Response:
        return ResponseNo(tag, self.message, self.code)


class SearchNotAllowed(CommandNotAllowed):
    """The ``SEARCH`` command contained a search key that could not be
    executed by the mailbox.

    Args:
        key: The search key that failed.

    """

    def __init__(self, key: bytes = None) -> None:
        command = b'SEARCH ' + key if key else b'SEARCH'
        super().__init__(command + b' not allowed.',
                         ResponseCode.of(b'CANNOT'))


class InvalidAuth(ResponseError):
    """The ``LOGIN`` or ``AUTHENTICATE`` commands received credential that the
    IMAP backend has rejected.

    """

    def __init__(self) -> None:
        super().__init__('Invalid authentication credentials.')

    def get_response(self, tag: bytes) -> ResponseNo:
        return ResponseNo(tag, str(self).encode('ascii'),
                          ResponseCode.of(b'AUTHENTICATIONFAILED'))


class MailboxError(ResponseError):
    """Parent exception for errors related to a mailbox.

    Args:
        mailbox: The name of the mailbox.
        message: The response message for the error.
        code: Optional response code for the error.

    """

    def __init__(self, mailbox: Optional[str], message: bytes,
                 code: Optional[ResponseCode] = None) -> None:
        super().__init__()
        self.mailbox = mailbox
        self.message = message
        self.code = code

    def get_response(self, tag: bytes) -> ResponseNo:
        return ResponseNo(tag, self.message, self.code)


class MailboxAbort(MailboxError):
    """An unrecoverable failure has aborted an operation on the mailbox. The
    operation may or may not succeed again.

    """

    def __init__(self) -> None:
        super().__init__(None, b'Mailbox operation aborted.',
                         ResponseCode.of(b'INUSE'))

    @classmethod
    def assertFalse(cls, cond: Any) -> None:
        """Asserts that `cond` is False.

        Args:
            cond: The condition to check.

        Raises:
            :exc:`MailboxAbort`

        """
        if cond:
            raise cls()


class MailboxNotFound(MailboxError):
    """The requested mailbox was not found.

    Args:
        mailbox: The name of the mailbox
        try_create: True if creating the mailbox first may help.

    """

    def __init__(self, mailbox: str, try_create: bool = False) -> None:
        code = ResponseCode.of(b'TRYCREATE' if try_create else b'NONEXISTENT')
        super().__init__(mailbox, b'Mailbox does not exist.', code)


class MailboxConflict(MailboxError):
    """The mailbox cannot be created or renamed because of a naming conflict
    with another mailbox.

    Args:
        mailbox: The name of the mailbox.

    """

    def __init__(self, mailbox: str) -> None:
        super().__init__(mailbox, b'Mailbox already exists.',
                         ResponseCode.of(b'ALREADYEXISTS'))


class MailboxHasChildren(MailboxError):
    """The mailbox cannot be deleted because there are other inferior
    hierarchical mailboxes below it.

    Args:
        mailbox: The name of the mailbox.

    """

    def __init__(self, mailbox: str) -> None:
        super().__init__(mailbox, b'Mailbox has inferior hierarchical names.')


class MailboxReadOnly(MailboxError):
    """The mailbox is opened read-only and the requested operation is not
    allowed.

    Args:
        mailbox: The name of the mailbox.

    """

    def __init__(self, mailbox: Optional[str] = None) -> None:
        super().__init__(mailbox, b'Mailbox is read-only.',
                         ResponseCode.of(b'READ-ONLY'))


class AppendFailure(MailboxError):
    """The mailbox append operation failed."""




import errno
import os
import os.path
from datetime import datetime
from mailbox import Maildir as _Maildir, MaildirMessage  # type: ignore
from typing import Dict, Optional, FrozenSet, Iterable, AsyncIterable

from pymap.concurrent import ReadWriteLock
from pymap.context import subsystem
from pymap.exceptions import MailboxNotFound, MailboxConflict, \
    MailboxHasChildren, NotSupportedError
from pymap.flags import FlagOp
from pymap.interfaces.message import AppendMessage, CachedMessage
from pymap.listtree import ListTree
from pymap.mailbox import MailboxSnapshot
from pymap.parsing.specials import FetchRequirement
from pymap.parsing.specials.flag import Flag, Seen
from pymap.selected import SelectedSet, SelectedMailbox

from .flags import MaildirFlags
from .io import NoChanges
from .layout import MaildirLayout
from .subscriptions import Subscriptions
from .uidlist import Record, UidList
from ..mailbox import MailboxDataInterface, MailboxSetInterface, \
    Message as _Message

__all__ = ['Maildir', 'Message', 'MailboxData', 'MailboxSet']


class Maildir(_Maildir):

    def claim_new(self) -> Iterable[str]:
        """Checks for messages in the ``new`` subdirectory, moving them to
        ``cur`` and returning their keys.

        """
        new_subdir = self._paths['new']
        cur_subdir = self._paths['cur']
        for name in os.listdir(new_subdir):
            new_path = os.path.join(new_subdir, name)
            cur_path = os.path.join(cur_subdir, name)
            try:
                os.rename(new_path, cur_path)
            except FileNotFoundError:
                pass
            else:
                yield name.rsplit(self.colon, 1)[0]

    def get_message_metadata(self, key: str) -> MaildirMessage:
        """Like :meth:`~mailbox.Maildir.get_message` but the message contents
        are not read from disk.

        """
        msg = MaildirMessage()
        subpath = self._lookup(key)
        subdir, name = os.path.split(subpath)
        msg.set_subdir(subdir)
        if self.colon in name:
            msg.set_info(name.rsplit(self.colon, 1)[-1])
        msg.set_date(os.path.getmtime(os.path.join(self._path, subpath)))
        return msg

    def update_metadata(self, key: str, msg: MaildirMessage) -> None:
        """Uses :func:`os.rename` to atomically update the message filename
        based on :meth:`~mailbox.MaildirMessage.get_info`.

        """
        subpath = self._lookup(key)
        subdir, name = os.path.split(subpath)
        new_subdir = msg.get_subdir()
        new_name = key + self.colon + msg.get_info()
        if subdir != new_subdir:
            raise ValueError('Message subdir may not be updated')
        elif name != new_name:
            new_subpath = os.path.join(msg.get_subdir(), new_name)
            old_path = os.path.join(self._path, subpath)
            new_path = os.path.join(self._path, new_subpath)
            os.rename(old_path, new_path)
            self._toc[key] = new_subpath


class Message(_Message):

    @property
    def maildir_flags(self) -> MaildirFlags:
        return self._kwargs['maildir_flags']

    @property
    def maildir_msg(self) -> MaildirMessage:
        flag_str = self.maildir_flags.to_maildir(self.permanent_flags)
        msg_bytes = bytes(self.get_body(binary=True))
        maildir_msg = MaildirMessage(msg_bytes)
        maildir_msg.set_flags(flag_str)
        maildir_msg.set_subdir('new' if self.recent else 'cur')
        if self.internal_date is not None:
            maildir_msg.set_date(self.internal_date.timestamp())
        return maildir_msg

    @classmethod
    def from_maildir(cls, uid: int, maildir_msg: MaildirMessage,
                     maildir_flags: 'MaildirFlags',
                     metadata_only: bool) -> 'Message':
        flag_set = maildir_flags.from_maildir(maildir_msg.get_flags())
        recent = maildir_msg.get_subdir() == 'new'
        msg_dt = datetime.fromtimestamp(maildir_msg.get_date())
        if metadata_only:
            return cls(uid, flag_set, msg_dt, recent=recent,
                       maildir_flags=maildir_flags)
        else:
            msg_data = bytes(maildir_msg)
            return cls.parse(uid, msg_data, flag_set, msg_dt, recent=recent,
                             maildir_flags=maildir_flags)


class MailboxData(MailboxDataInterface[Message]):

    db_retry_count = 100
    db_retry_delay = 0.1
    filename_db = '.uid'
    filename_tmp_db = 'tmp.uid'

    def __init__(self, guid: bytes, maildir: Maildir, path: str) -> None:
        super().__init__()
        self._guid = guid
        self._maildir = maildir
        self._path = path
        self._uid_validity = 0
        self._next_uid = 0
        self._flags: Optional[MaildirFlags] = None
        self._messages_lock = subsystem.get().new_rwlock()
        self._selected_set = SelectedSet()

    @property
    def guid(self) -> bytes:
        return self._guid

    @property
    def readonly(self) -> bool:
        return False

    @property
    def uid_validity(self) -> int:
        return self._uid_validity

    @property
    def maildir_flags(self) -> MaildirFlags:
        if self._flags is not None:
            return self._flags
        self._flags = flags = MaildirFlags.file_read(self._path)
        return flags

    @property
    def permanent_flags(self) -> FrozenSet[Flag]:
        return self.maildir_flags.permanent_flags

    @property
    def messages_lock(self) -> ReadWriteLock:
        return self._messages_lock

    @property
    def selected_set(self) -> SelectedSet:
        return self._selected_set

    async def update_selected(self, selected: SelectedMailbox) \
            -> SelectedMailbox:
        all_messages = [msg async for msg in self.messages()]
        selected.set_messages(all_messages)
        return selected

    async def add(self, append_msg: AppendMessage, recent: bool = False) \
            -> Message:
        message = Message.parse(0, append_msg.message, append_msg.flag_set,
                                append_msg.when, recent=True,
                                maildir_flags=self.maildir_flags)
        async with self.messages_lock.write_lock():
            maildir_msg = message.maildir_msg
            if recent:
                maildir_msg.set_subdir('new')
            key = self._maildir.add(maildir_msg)
            filename = key + ':' + maildir_msg.get_info()
        async with UidList.with_write(self._path) as uidl:
            new_rec = Record(uidl.next_uid, {}, filename)
            uidl.next_uid += 1
            uidl.set(new_rec)
        msg_copy = message.copy(new_rec.uid)
        msg_copy.recent = recent
        return msg_copy

    async def get(self, uid: int, cached_msg: CachedMessage = None,
                  requirement: FetchRequirement = FetchRequirement.METADATA) \
            -> Optional[Message]:
        async with UidList.with_read(self._path) as uidl:
            next_uid = uidl.next_uid
            if uid < 1 or uid >= next_uid:
                raise IndexError(uid)
            try:
                key = uidl.get(uid).key
            except KeyError:
                if cached_msg is not None:
                    return Message(cached_msg.uid, cached_msg.permanent_flags,
                                   cached_msg.internal_date, expunged=True)
                else:
                    return None
        metadata_only = (requirement == FetchRequirement.METADATA)
        async with self.messages_lock.read_lock():
            try:
                if metadata_only:
                    maildir_msg = self._maildir.get_message_metadata(key)
                else:
                    maildir_msg = self._maildir.get_message(key)
            except (KeyError, FileNotFoundError):
                if cached_msg is not None:
                    return Message(cached_msg.uid, cached_msg.permanent_flags,
                                   cached_msg.internal_date, expunged=True)
                else:
                    return None
            return Message.from_maildir(uid, maildir_msg, self.maildir_flags,
                                        metadata_only)

    async def delete(self, uids: Iterable[int]) -> None:
        async with UidList.with_read(self._path) as uidl:
            records = uidl.get_all(uids)
        async with self.messages_lock.write_lock():
            for uid, rec in records.items():
                try:
                    self._maildir.remove(rec.key)
                except (KeyError, FileNotFoundError):
                    pass

    async def claim_recent(self, selected: SelectedMailbox) -> None:
        async with self.messages_lock.write_lock():
            keys = self._maildir.claim_new()
        async with UidList.with_read(self._path) as uidl:
            for rec in uidl.records:
                if rec.key in keys:
                    selected.session_flags.add_recent(rec.uid)

    async def update_flags(self, messages: Iterable[Message],
                           flag_set: FrozenSet[Flag], mode: FlagOp) -> None:
        msgs_map = {msg.uid: msg for msg in messages}
        async with UidList.with_read(self._path) as uidl:
            records = uidl.get_all(msgs_map.keys())
        async with self.messages_lock.write_lock():
            for uid, rec in records.items():
                key = rec.key
                msg = msgs_map[uid]
                msg.permanent_flags = mode.apply(msg.permanent_flags, flag_set)
                flag_str = self.maildir_flags.to_maildir(msg.permanent_flags)
                try:
                    maildir_msg = self._maildir.get_message_metadata(key)
                except (KeyError, FileNotFoundError):
                    continue
                maildir_msg.set_flags(flag_str)
                maildir_msg.set_subdir('new' if msg.recent else 'cur')
                try:
                    self._maildir.update_metadata(key, maildir_msg)
                except (KeyError, FileNotFoundError):
                    pass

    async def cleanup(self) -> None:
        self._maildir.clean()
        keys = await self._get_keys()
        async with UidList.with_write(self._path) as uidl:
            for rec in list(uidl.records):
                key = rec.key
                info = keys.get(key)
                if info is None:
                    uidl.remove(rec.uid)
                else:
                    filename = key + ':' + info
                    new_rec = Record(rec.uid, rec.fields, filename)
                    uidl.set(new_rec)

    async def messages(self) -> AsyncIterable[Message]:
        async with UidList.with_read(self._path) as uidl:
            uids = {rec.uid: rec.key for rec in uidl.records}
        async with self.messages_lock.read_lock():
            for uid, key in uids.items():
                try:
                    maildir_msg = self._maildir.get_message_metadata(key)
                except (KeyError, FileNotFoundError):
                    pass
                else:
                    yield Message.from_maildir(
                        uid, maildir_msg, self.maildir_flags, True)

    async def reset(self) -> 'MailboxData':
        keys = await self._get_keys()
        async with UidList.with_write(self._path) as uidl:
            for rec in uidl.records:
                keys.pop(rec.key, None)
            if not keys:
                raise NoChanges()
            for key, info in keys.items():
                filename = key + ':' + info
                new_rec = Record(uidl.next_uid, {}, filename)
                uidl.next_uid += 1
                uidl.set(new_rec)
        self._uid_validity = uidl.uid_validity
        self._next_uid = uidl.next_uid
        return self

    async def snapshot(self) -> MailboxSnapshot:
        exists = 0
        recent = 0
        unseen = 0
        first_unseen: Optional[int] = None
        next_uid = self._next_uid
        async for msg in self.messages():
            exists += 1
            if msg.recent:
                recent += 1
            if Seen not in msg.permanent_flags:
                unseen += 1
                if first_unseen is None:
                    first_unseen = exists
        return MailboxSnapshot(self.guid, self.readonly, self.uid_validity,
                               self.permanent_flags, self.session_flags,
                               exists, recent, unseen, first_unseen, next_uid)

    async def _get_keys(self) -> Dict[str, str]:
        keys: Dict[str, str] = {}
        async with self.messages_lock.read_lock():
            for key in self._maildir.keys():
                try:
                    msg = self._maildir.get_message_metadata(key)
                except (KeyError, FileNotFoundError):
                    pass
                else:
                    keys[key] = msg.get_info()
        return keys


class MailboxSet(MailboxSetInterface[MailboxData]):

    def __init__(self, maildir: Maildir, layout: MaildirLayout) -> None:
        super().__init__()
        self._layout = layout
        self._inbox_maildir = maildir
        self._path = layout.path
        self._cache: Dict[str, 'MailboxData'] = {}

    @property
    def delimiter(self) -> str:
        return '/'

    async def set_subscribed(self, name: str, subscribed: bool) -> None:
        async with Subscriptions.with_write(self._path) as subs:
            subs.set(name, subscribed)

    async def list_subscribed(self) -> ListTree:
        async with Subscriptions.with_read(self._path) as subs:
            subscribed = frozenset(subs.subscribed)
        mailboxes = [name for name in self._layout.list_folders(self.delimiter)
                     if name in subscribed]
        return ListTree(self.delimiter).update('INBOX', *mailboxes)

    async def list_mailboxes(self) -> ListTree:
        mailboxes = self._layout.list_folders(self.delimiter)
        return ListTree(self.delimiter).update('INBOX', *mailboxes)

    async def get_mailbox(self, name: str,
                          try_create: bool = False) -> 'MailboxData':
        if name == 'INBOX':
            maildir = self._inbox_maildir
        else:
            try:
                maildir = self._layout.get_folder(name, self.delimiter)
            except FileNotFoundError:
                raise MailboxNotFound(name, try_create)
        if name in self._cache:
            mbx = self._cache[name]
        else:
            path = self._layout.get_path(name, self.delimiter)
            async with UidList.with_open(path) as uidl:
                guid = uidl.global_uid
            mbx = MailboxData(guid, maildir, path)
            self._cache[name] = mbx
        return await mbx.reset()

    async def add_mailbox(self, name: str) -> None:
        try:
            self._layout.add_folder(name, self.delimiter)
        except FileExistsError:
            raise MailboxConflict(name)

    async def delete_mailbox(self, name: str) -> None:
        try:
            self._layout.remove_folder(name, self.delimiter)
        except FileNotFoundError:
            raise MailboxNotFound(name)
        except OSError as exc:
            if exc.errno == errno.ENOTEMPTY:
                raise MailboxHasChildren(name) from exc
            raise exc

    async def rename_mailbox(self, before: str, after: str) -> None:
        if before == 'INBOX':
            raise NotSupportedError()  # TODO
        else:
            self._layout.rename_folder(before, after, self.delimiter)




import asyncio
import logging
import re
from argparse import ArgumentParser
from asyncio import Task, StreamReader, StreamWriter, AbstractServer
from base64 import b64encode, b64decode
from collections import OrderedDict
from typing import Optional, Union, Mapping, Dict, List

from pymap import __version__
from pymap.bytes import BytesFormat
from pymap.config import IMAPConfig
from pymap.context import socket_info, language_code
from pymap.exceptions import InvalidAuth
from pymap.interfaces.backend import BackendInterface, ServiceInterface
from pymap.interfaces.session import LoginProtocol, SessionInterface
from pymap.parsing.exceptions import NotParseable
from pymap.parsing.primitives import String
from pymap.sockets import SocketInfo
from pysasl import ServerChallenge, AuthenticationError

from .command import Command, NoOpCommand, LogoutCommand, CapabilityCommand, \
    AuthenticateCommand, UnauthenticateCommand, StartTLSCommand
from .response import Condition, Response, BadCommandResponse, NoOpResponse, \
    CapabilitiesResponse
from .state import FilterState

__all__ = ['ManageSieveService', 'ManageSieveServer', 'ManageSieveConnection']

_log = logging.getLogger(__name__)


class ManageSieveService(ServiceInterface):
    """A pymap service that implements a ManageSieve server to control the
    sieve scripts associated with a login.

    See Also:
        `RFC 5804 <https://tools.ietf.org/html/rfc5804>`_,
        :class:`~pymap.interfaces.filter.FilterSetInterface`

    """

    def __init__(self, server: AbstractServer) -> None:
        super().__init__()
        self._server = server
        self._task = asyncio.create_task(self._run())

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        group = parser.add_argument_group('managesieve service')
        group.add_argument('--sieve-host', action='append', metavar='IFACE',
                           help='the network interface to listen on')
        group.add_argument('--sieve-port',  metavar='NUM', default='sieve',
                           help='the port or service name to listen on')

    @classmethod
    async def start(cls, backend: BackendInterface,
                    config: IMAPConfig) -> 'ManageSieveService':
        managesieve_server = ManageSieveServer(backend.login, backend.config)
        host: Optional[str] = config.args.sieve_host
        port: Union[str, int] = config.args.sieve_port
        server = await asyncio.start_server(
            managesieve_server, host=host, port=port)
        return cls(server)

    @property
    def task(self) -> Task:
        return self._task

    async def _run(self) -> None:
        async with self._server:  # type: ignore
            await self._server.serve_forever()  # type: ignore


class ManageSieveServer:
    """Callable object that creates and runs :class:`ManageSieveConnection`
    objects when :func:`asyncio.start_server` receives a new connection.

    Args:
        login: Login callback that takes authentication credentials and returns
            a :class:`~pymap.interfaces.session.SessionInterface` object.

    """

    def __init__(self, login: LoginProtocol, config: IMAPConfig) -> None:
        super().__init__()
        self._login = login
        self._config = config

    async def __call__(self, reader: StreamReader,
                       writer: StreamWriter) -> None:
        conn = ManageSieveConnection(self._config, reader, writer)
        try:
            await conn.run(self._login)
        finally:
            writer.close()


class ManageSieveConnection:
    """Runs a single ManageSieve connection from start to finish.

    Args:
        reader: The input stream for the socket.
        writer: The output stream for the socket.

    """

    _lines = re.compile(br'\r?\n')
    _literal_plus = re.compile(br'{(\d+)\+}\r?\n$')
    _impl = b'pymap managesieve ' + __version__.encode('ascii')

    def __init__(self, config: IMAPConfig, reader: StreamReader,
                 writer: StreamWriter) -> None:
        super().__init__()
        self.config = config
        self.auth = config.initial_auth
        self.params = config.parsing_params.copy(allow_continuations=False)
        self._offer_starttls = b'STARTTLS' in config.initial_capability
        self._session: Optional[SessionInterface] = None
        self._reset_streams(reader, writer)

    def _reset_streams(self, reader: StreamReader,
                       writer: StreamWriter) -> None:
        self.reader = reader
        self.writer = writer
        socket_info.set(SocketInfo(writer))

    @property
    def capabilities(self) -> Mapping[bytes, Optional[bytes]]:
        ret: Dict[bytes, Optional[bytes]] = OrderedDict()
        ret[b'IMPLEMENTATION'] = self._impl
        if self._session is None:
            ret[b'SASL'] = b' '.join(
                mech.name for mech in self.auth.server_mechanisms)
        ret[b'SIEVE'] = b''
        if self._offer_starttls and self._session is None:
            ret[b'STARTTLS'] = None
        try:
            ret[b'LANGUAGE'] = language_code.get().encode('ascii')
        except LookupError:
            pass
        if self._session is not None:
            ret[b'OWNER'] = self._session.owner.encode('utf-8')
        ret[b'UNAUTHENTICATE'] = None
        ret[b'VERSION'] = b'1.0'
        return ret

    @classmethod
    def _print(cls, log_format: str, output: bytes) -> None:
        if _log.isEnabledFor(logging.DEBUG):
            fd = socket_info.get().socket.fileno()
            lines = cls._lines.split(output)
            if not lines[-1]:
                lines = lines[:-1]
            for line in lines:
                line_str = str(line, 'utf-8', 'replace')
                _log.debug(log_format, fd, line_str)

    async def _read_data(self) -> memoryview:
        data = bytearray()
        while True:
            data += await self.reader.readline()
            if not data.endswith(b'\n'):
                raise EOFError()
            match = self._literal_plus.search(data)
            if not match:
                break
            literal_length = int(match.group(1))
            data += await self.reader.readexactly(literal_length)
        self._print('%d -->| %s', data)
        return memoryview(data)

    async def _read_command(self) -> Command:
        cmd_buf = await self._read_data()
        cmd, _ = Command.parse(cmd_buf, self.params)
        return cmd

    async def _write_response(self, resp: Response) -> None:
        resp.write(self.writer)
        try:
            await self.writer.drain()
        except ConnectionError:
            pass
        else:
            self._print('%d <--| %s', bytes(resp))

    async def _do_greeting(self, login: LoginProtocol) -> None:
        preauth_creds = self.config.preauth_credentials
        if preauth_creds:
            self._session = await login(preauth_creds, self.config)
        greeting = CapabilitiesResponse(self.capabilities)
        await self._write_response(greeting)

    async def _do_authenticate(self, login: LoginProtocol,
                               cmd: AuthenticateCommand) -> Response:
        mech = self.auth.get_server(cmd.mech_name)
        if not mech:
            return Response(Condition.NO, text='Invalid SASL mechanism.')
        responses: List[ServerChallenge] = []
        if cmd.initial_data is not None:
            chal = ServerChallenge(b'')
            chal.set_response(b64decode(cmd.initial_data))
            responses.append(chal)
        while True:
            try:
                creds, final = mech.server_attempt(responses)
            except ServerChallenge as chal:
                chal_bytes = b64encode(chal.get_challenge())
                chal_str = String.build(chal_bytes)
                chal_str.write(self.writer)
                self.writer.write(b'\r\n')
                await self.writer.drain()
                resp_bytes = await self._read_data()
                resp_str, _ = String.parse(resp_bytes, self.params)
                chal.set_response(b64decode(resp_str.value))
                if resp_str.value == b'*':
                    raise AuthenticationError('Authentication cancelled.')
                responses.append(chal)
            except AuthenticationError as exc:
                return Response(Condition.NO, text=str(exc))
            else:
                break
        if final is None:
            code: Optional[bytes] = None
        else:
            code = BytesFormat(b'SASL %b') % String.build(final)
        try:
            session = await login(creds, self.config)
        except InvalidAuth as exc:
            return Response(Condition.NO, text=str(exc))
        else:
            if session.filter_set is None:
                return Response(Condition.NO, text='Filters not supported.')
            self._session = session
            return Response(Condition.OK, code=code)

    async def _do_unauthenticate(self) -> Response:
        if self._session is None:
            return Response(Condition.NO, text='Not authenticated.')
        else:
            self._session = None
            return Response(Condition.OK)

    async def _do_starttls(self) -> Response:
        ssl_context = self.config.ssl_context
        if ssl_context is None:
            raise ValueError('ssl_context is None')
        elif not self._offer_starttls:
            return Response(Condition.NO, text='Bad command.')
        resp = Response(Condition.OK)
        await self._write_response(resp)
        loop = asyncio.get_event_loop()
        transport = self.writer.transport
        protocol = transport.get_protocol()  # type: ignore
        new_transport = await loop.start_tls(  # type: ignore
            transport, protocol, ssl_context, server_side=True)
        protocol._stream_reader = StreamReader(loop=loop)
        protocol._client_connected_cb = self._reset_streams
        protocol.connection_made(new_transport)
        self._print('%d <->| %s', b'<TLS handshake>')
        self._offer_starttls = False
        self.auth = self.config.insecure_auth
        return CapabilitiesResponse(self.capabilities)

    async def run(self, login: LoginProtocol):
        """Start the socket communication with the server greeting, and then
        enter the command/response cycle.

        Args:
            login: The login/authentication function.

        """
        self._print('%d +++| %s', bytes(socket_info.get()))
        await self._do_greeting(login)
        while True:
            resp: Response
            try:
                cmd = await self._read_command()
            except (ConnectionError, EOFError):
                break
            except NotParseable as exc:
                resp = BadCommandResponse(exc)
            else:
                try:
                    if isinstance(cmd, NoOpCommand):
                        resp = NoOpResponse(cmd.tag)
                    elif isinstance(cmd, LogoutCommand):
                        resp = Response(Condition.BYE)
                    elif isinstance(cmd, CapabilityCommand):
                        resp = CapabilitiesResponse(self.capabilities)
                    elif self._session is None:
                        if isinstance(cmd, AuthenticateCommand):
                            resp = await self._do_authenticate(login, cmd)
                        elif isinstance(cmd, StartTLSCommand):
                            resp = await self._do_starttls()
                        else:
                            resp = Response(Condition.NO, text='Bad command.')
                    else:
                        if isinstance(cmd, UnauthenticateCommand):
                            resp = await self._do_unauthenticate()
                        else:
                            assert self._session.filter_set is not None
                            state = FilterState(self._session.filter_set,
                                                self.config)
                            resp = await state.run(cmd)
                except Exception:
                    _log.exception('Unhandled exception')
                    resp = Response(Condition.NO, text='Server error.')
            await self._write_response(resp)
            if resp.is_bye:
                break
        self._print('%d ---| %s', b'<disconnected>')




from abc import abstractmethod
from typing import Generic, Type, Optional, Tuple, Sequence

from pkg_resources import iter_entry_points, DistributionNotFound

from .interfaces.filter import FilterValueT, FilterInterface, \
    FilterSetInterface

__all__ = ['FilterCompiler', 'EntryPointFilterSet', 'SingleFilterSet']


class FilterCompiler(Generic[FilterValueT]):
    """Abstract base class for classes which can compile a filter value into an
    implementation.

    """

    @abstractmethod
    async def compile(self, value: FilterValueT) -> FilterInterface:
        """Compile the filter value and return the resulting implementation.

        Args:
            value: The filter value.

        """
        ...


class EntryPointFilterSet(FilterSetInterface[FilterValueT]):
    """Base class for filter set implementations that use a filter compiler
    declared in the ``pymap.filter`` entry point. The declared entry points
    must sub-class :class:`FilterCompiler`.

    Args:
        entry_point: The entry point name.
        group: The entry point group.

    """

    def __init__(self, entry_point: str, *,
                 group: str = 'pymap.filter') -> None:
        super().__init__()
        self._group = group
        self._entry_point = entry_point
        self._filter_compiler: Optional[FilterCompiler] = None

    @property
    def filter_compiler(self) -> FilterCompiler:
        if self._filter_compiler is None:
            filter_cls: Optional[Type[FilterCompiler]] = None
            name = self._entry_point
            for entry_point in iter_entry_points(self._group, name):
                try:
                    filter_cls = entry_point.load()
                except DistributionNotFound:
                    pass  # optional dependencies not installed
            if filter_cls is None:
                raise LookupError(f'{self._group}:{name}')
            self._filter_compiler = filter_cls()
        return self._filter_compiler

    async def compile(self, value: FilterValueT) -> FilterInterface:
        return await self.filter_compiler.compile(value)


class SingleFilterSet(FilterSetInterface[FilterValueT]):
    """Base class for a filter set that does not use named filter
    implementations, it contains only a single active filter implementation.

    """

    @property
    def name(self) -> str:
        """The permanent name for the active filter."""
        return 'active'

    async def put(self, name: str, value: FilterValueT) -> None:
        if name == self.name:
            await self.replace_active(value)

    async def delete(self, name: str) -> None:
        if name == self.name:
            await self.replace_active(None)
        else:
            raise KeyError(name)

    async def rename(self, before_name: str, after_name: str) -> None:
        raise NotImplementedError()

    async def set_active(self, name: Optional[str]) -> None:
        if name is None:
            await self.replace_active(None)
        elif name != self.name:
            raise KeyError(name)

    async def get(self, name: str) -> FilterValueT:
        if name == self.name:
            value = await self.get_active()
            if value is None:
                raise KeyError(name)
            else:
                return value
        else:
            raise KeyError(name)

    async def get_all(self) -> Tuple[Optional[str], Sequence[str]]:
        value = await self.get_active()
        if value is None:
            return None, []
        else:
            return self.name, [self.name]

    async def replace_active(self, value: Optional[FilterValueT]) -> None:
        """Replace the current active filter value with a new value.

        Args:
            value: The new filter value.

        """
        raise NotImplementedError()

    @abstractmethod
    async def get_active(self) -> Optional[FilterValueT]:
        ...




from abc import abstractmethod, ABCMeta
from email.headerregistry import ContentTypeHeader
from email.policy import SMTP
from itertools import chain, islice
from typing import Optional, Tuple, Iterable, Sequence, Mapping, List, Dict
from typing_extensions import Final

from .parsed import ParsedHeaders
from ._util import whitespace, find_any, get_raw
from ..bytes import WriteStream, Writeable

__all__ = ['MessageContent', 'MessageHeader', 'MessageBody']

_default_type: Final[ContentTypeHeader] = (  # type: ignore
    SMTP.header_fetch_parse('Content-Type', 'text/plain'))

_Line = Tuple[int, int, int]
_Lines = Sequence[_Line]
_Headers = Mapping[bytes, Sequence[Sequence[memoryview]]]
_Folded = Sequence[Tuple[bytes, memoryview]]


class MessageContent(Writeable):
    """Contains the message content, parsed for IMAP processing.

    Attributes:
        lines: The number of lines in the message content.
        header: The message header.
        body: The message body.

    """

    __slots__ = ['_raw', 'lines', 'header', 'body']

    def __init__(self, raw: Sequence[memoryview], lines: int,
                 header: 'MessageHeader', body: 'MessageBody') -> None:
        super().__init__()
        self._raw: Final = raw
        self.lines: Final = lines
        self.header: Final = header
        self.body: Final = body

    def walk(self) -> Iterable['MessageContent']:
        """Iterate through the message and all its nested sub-parts in the
        order they occur.

        """
        if self.body.has_nested:
            return chain([self], *(part.walk() for part in self.body.nested))
        else:
            return [self]

    @property
    def is_rfc822(self) -> bool:
        """True if the content-type of the message is ``message/rfc822``."""
        ct_hdr = self.header.parsed.content_type
        if ct_hdr is None:
            return False
        else:
            return ct_hdr.content_type == 'message/rfc822'

    @classmethod
    def parse(cls, data: bytes) -> 'MessageContent':
        """Parse the bytestring into message content.

        Args:
            data: The bytestring to parse.

        """
        lines = cls._find_lines(data)
        view = memoryview(data)
        return cls._parse(data, view, lines)

    @classmethod
    def parse_split(cls, header: bytes, body: bytes) -> 'MessageContent':
        """Parse the header and body bytestrings into message content.

        Args:
            header: The header bytestring to parse.
            body: The body bytestring to parse.

        """
        header_lines = cls._find_lines(header)
        body_lines = cls._find_lines(body)
        header_view = memoryview(header)
        body_view = memoryview(body)
        return cls._parse_split([header_view, body_view], header, body,
                                header_view, body_view,
                                header_lines, body_lines)

    @classmethod
    def _parse(cls, data: bytes, view: memoryview, lines: _Lines) \
            -> 'MessageContent':
        header_lines, body_lines = cls._split_lines(data, lines)
        raw = get_raw(view, lines)
        return cls._parse_split([raw], data, data, view, view,
                                header_lines, body_lines)

    @classmethod
    def _parse_split(cls, raw: Sequence[memoryview],
                     header_data: bytes, body_data: bytes,
                     header_view: memoryview, body_view: memoryview,
                     header_lines: _Lines, body_lines: _Lines) \
            -> 'MessageContent':
        header = MessageHeader._parse(header_data, header_view, header_lines)
        content_type = header.parsed.content_type
        body = MessageBody._parse(body_data, body_view, body_lines,
                                  content_type)
        num_lines = len(header_lines) + len(body_lines) - 1
        return cls(raw, num_lines, header, body)

    @classmethod
    def _find_lines(cls, data: bytes) -> _Lines:
        start = 0
        end = len(data)
        ret: List[_Line] = []
        while True:
            idx = data.find(b'\n', start, end)
            if idx < 0:
                ret.append((start, end, end))
                break
            next_start = idx + 1
            idx_minus_1 = idx - 1
            if idx_minus_1 >= start and data[idx_minus_1] == 0x0d:
                idx = idx_minus_1
            ret.append((start, idx, next_start))
            start = next_start
        return ret

    @classmethod
    def _split_lines(cls, data: bytes, lines: _Lines) -> Tuple[_Lines, _Lines]:
        for i, line in enumerate(lines):
            start, end, _ = line
            ws_end = find_any(data, whitespace, start, end, False, False)
            if ws_end < 0:
                return lines[0:i + 1], lines[i + 1:]
        return [], lines

    def write(self, writer: WriteStream) -> None:
        """Write the object to the stream, with one or more calls to
        :meth:`~pymap.bytes.WriteStream.write`.

        Args:
            writer: The output stream.

        """
        for part in self._raw:
            writer.write(bytes(part))

    def __len__(self) -> int:
        return sum(len(part) for part in self._raw)

    def __bytes__(self) -> bytes:
        return b''.join(bytes(part) for part in self._raw)


class MessageHeader(Writeable):
    """The message header. Contains lines in the form of ``Header: value\\n``,
    possibly folded onto multiple lines where subsequent lines start with
    whitespace.

    Attributes:
        lines: The number of lines in the message header.
        folded: A list of headers, as they occurred in the original data, as
            tuples of the lower-cased header name and the full header value,
            including the header name and any extra folded lines.
        parsed: The message headers, as a dictionary-like object that parses
            headers on-demand.

    """

    __slots__ = ['_raw', 'lines', 'folded', 'parsed']

    def __init__(self, raw: Sequence[memoryview], lines: int, folded: _Folded,
                 parsed: ParsedHeaders) -> None:
        super().__init__()
        self._raw: Final = raw
        self.lines: Final = lines
        self.folded: Final = folded
        self.parsed: Final = parsed

    @classmethod
    def _parse(cls, data: bytes, view: memoryview,
               lines: _Lines) -> 'MessageHeader':
        folds = cls._find_folds(data, lines)
        folded, parsed = cls._parse_headers(data, view, folds)
        raw = get_raw(view, lines)
        return cls([raw], len(lines), folded, parsed)

    @classmethod
    def _find_folds(cls, data: bytes, lines: _Lines) -> Sequence[_Lines]:
        ret: List[List[Tuple[int, int, int]]] = []
        if not lines:
            return []
        for line in islice(lines, len(lines) - 1):
            start, end, _ = line
            length = end - start
            if length >= 1 and data[start] in whitespace:
                if ret:
                    ret[-1].append(line)
            else:
                ret.append([line])
        return ret

    @classmethod
    def _parse_headers(cls, data: bytes, view: memoryview,
                       folds: Sequence[_Lines]) \
            -> Tuple[_Folded, ParsedHeaders]:
        headers: Dict[bytes, List[List[bytes]]] = {}
        folded: List[Tuple[bytes, memoryview]] = []
        for group in folds:
            start, end, _ = group[0]
            colon = data.find(b':', start, end)
            if colon < 0:
                continue
            name = data[start:colon].strip().lower()
            values = headers.setdefault(name, [])
            values.append([data[start:end] for start, end, _ in group])
            folded.append((name, get_raw(view, group)))
        return folded, ParsedHeaders(headers)

    @classmethod
    def _join_group(cls, data: bytes, group: _Lines, colon: int) -> bytes:
        first_start, first_end, _ = group[0]
        value_offset = colon - first_start + 1
        length = sum(end - start for start, end, _ in group)
        ret = bytearray(length - value_offset)
        current = first_end - value_offset
        ret[0:current] = data[colon + 1:first_end]
        for start, end, _ in islice(group, 1, None):
            current_end = current + (end - start)
            ret[current:current_end] = data[start:end]
            current = current_end
        return bytes(ret)

    def write(self, writer: WriteStream) -> None:
        """Write the object to the stream, with one or more calls to
        :meth:`~pymap.bytes.WriteStream.write`.

        Args:
            writer: The output stream.

        """
        for part in self._raw:
            writer.write(bytes(part))

    def __len__(self) -> int:
        return sum(len(part) for part in self._raw)

    def __bytes__(self) -> bytes:
        return b''.join(bytes(part) for part in self._raw)


class MessageBody(Writeable, metaclass=ABCMeta):
    """The message body, starting immediately after the header. The body may
    contain nested sub-parts, which are each valid :class:`MessageContent`
    objects.

    Attributes:
        lines: The number of lines in the message body.
        content_type: The content type of the message body.

    """

    __slots__ = ['_raw', 'lines', 'content_type']

    def __init__(self, raw: Sequence[memoryview], lines: int,
                 content_type: ContentTypeHeader) -> None:
        super().__init__()
        self._raw: Final = raw
        self.lines: Final = lines
        self.content_type: Final = content_type

    @property
    @abstractmethod
    def has_nested(self) -> bool:
        """True if the message body is composed of nested sub-parts."""
        ...

    @property
    @abstractmethod
    def nested(self) -> Sequence[MessageContent]:
        """If :attr:`.has_nested` is True, contains the list of sub-parts."""
        ...

    @classmethod
    def _parse(cls, data: bytes, view: memoryview, lines: _Lines,
               content_type: Optional[ContentTypeHeader]) -> 'MessageBody':
        if content_type is None:
            content_type = _default_type
        maintype = content_type.maintype
        if maintype == 'multipart':
            boundary = cls._get_boundary(content_type)
            if boundary:
                return _MultipartBody._parse_body(
                    data, view, lines, content_type, boundary)
        elif maintype == 'message' and content_type.subtype == 'rfc822':
            return _RFC822Body._parse_body(data, view, lines, content_type)
        return _SinglepartBody._parse_body(data, view, lines, content_type)

    @classmethod
    def _get_boundary(cls, content_type: ContentTypeHeader) -> Optional[bytes]:
        try:
            boundary = content_type.params['boundary']
        except KeyError:
            pass
        else:
            if boundary:
                try:
                    return boundary.encode('ascii')
                except UnicodeError:
                    pass
        return None

    def write(self, writer: WriteStream) -> None:
        """Write the object to the stream, with one or more calls to
        :meth:`~pymap.bytes.WriteStream.write`.

        Args:
            writer: The output stream.

        """
        for part in self._raw:
            writer.write(bytes(part))

    def __len__(self) -> int:
        return sum(len(part) for part in self._raw)

    def __bytes__(self) -> bytes:
        return b''.join(bytes(part) for part in self._raw)


class _MultipartBody(MessageBody):

    __slots__ = ['_nested']

    def __init__(self, raw: Sequence[memoryview], lines: int,
                 content_type: ContentTypeHeader,
                 nested: Sequence[MessageContent]) -> None:
        super().__init__(raw, lines, content_type)
        self._nested = nested

    @property
    def has_nested(self) -> bool:
        return True

    @property
    def nested(self) -> Sequence[MessageContent]:
        return self._nested

    @classmethod
    def _parse_body(cls, data: bytes, view: memoryview, lines: _Lines,
                    content_type: ContentTypeHeader, boundary: bytes) \
            -> '_MultipartBody':
        parts = cls._find_parts(data, view, lines, boundary)
        nested: List[MessageContent] = []
        for part_lines in parts:
            sub_content = MessageContent._parse(data, view, part_lines)
            nested.append(sub_content)
        raw = get_raw(view, lines)
        return cls([raw], len(lines), content_type, nested)

    @classmethod
    def _find_parts(cls, data: bytes, view: memoryview, lines: _Lines,
                    boundary: bytes) -> Sequence[_Lines]:
        ret: List[List[_Line]] = []
        part_match = (b'--%s' % boundary, b'--%s' % boundary)
        stop_match = (b'--%s--' % boundary, b'--%s--' % boundary)
        for line in lines:
            start, end, _ = line
            line_view = view[start:end]
            if any(line_view == m for m in stop_match):
                break
            elif any(line_view == m for m in part_match):
                ret.append([])
            elif ret:
                ret[-1].append(line)
        return ret


class _RFC822Body(MessageBody):

    __slots__ = ['_nested']

    def __init__(self, subpart: MessageContent,
                 content_type: ContentTypeHeader) -> None:
        super().__init__(subpart._raw, subpart.lines, content_type)
        self._nested: Sequence[MessageContent] = [subpart]

    @property
    def has_nested(self) -> bool:
        return True

    @property
    def nested(self) -> Sequence[MessageContent]:
        return self._nested

    @classmethod
    def _parse_body(cls, data: bytes, view: memoryview, lines: _Lines,
                    content_type: ContentTypeHeader) -> '_RFC822Body':
        subpart = MessageContent._parse(data, view, lines)
        return cls(subpart, content_type)


class _SinglepartBody(MessageBody):

    def __init__(self, raw: Sequence[memoryview], lines: int,
                 content_type: ContentTypeHeader) -> None:
        super().__init__(raw, lines, content_type)

    @property
    def has_nested(self) -> bool:
        return False

    @property
    def nested(self) -> Sequence[MessageContent]:
        return []

    @classmethod
    def _parse_body(cls, data: bytes, view: memoryview, lines: _Lines,
                    content_type: ContentTypeHeader) -> '_SinglepartBody':
        raw = get_raw(view, lines)
        return cls([raw], len(lines), content_type)




from collections import OrderedDict
from typing import Optional, Dict, List, Callable, Union, Tuple, Awaitable, \
    Iterable

from pymap.bytes import MaybeBytes
from pymap.concurrent import Event
from pymap.config import IMAPConfig
from pymap.context import socket_info
from pymap.exceptions import CommandNotAllowed, CloseConnection
from pymap.fetch import MessageAttributes
from pymap.interfaces.session import SessionInterface, LoginProtocol
from pymap.parsing.command import CommandAuth, CommandNonAuth, CommandSelect, \
    Command
from pymap.parsing.command.any import CapabilityCommand, LogoutCommand, \
    NoOpCommand
from pymap.parsing.command.nonauth import AuthenticateCommand, LoginCommand, \
    StartTLSCommand
from pymap.parsing.command.auth import AppendCommand, CreateCommand, \
    DeleteCommand, ListCommand, RenameCommand, SelectCommand, StatusCommand, \
    SubscribeCommand, UnsubscribeCommand
from pymap.parsing.command.select import CheckCommand, CloseCommand, \
    IdleCommand, ExpungeCommand, CopyCommand, FetchCommand, StoreCommand, \
    SearchCommand
from pymap.parsing.commands import InvalidCommand
from pymap.parsing.primitives import ListP, Number
from pymap.parsing.response import Response, ResponseOk, ResponseNo, \
    ResponseBad,  ResponseCode, ResponsePreAuth
from pymap.parsing.response.code import Capability, PermanentFlags, UidNext, \
    UidValidity, Unseen
from pymap.parsing.response.specials import FlagsResponse, ExistsResponse, \
    RecentResponse, FetchResponse, ListResponse, LSubResponse, \
    SearchResponse, StatusResponse
from pymap.parsing.specials import FetchAttribute, StatusAttribute
from pymap.selected import SelectedMailbox
from pysasl import AuthenticationCredentials

__all__ = ['ConnectionState']

_AuthCommands = Union[AuthenticateCommand, LoginCommand]
_CommandFunc = Callable[[Command],
                        Awaitable[Tuple[Response, SelectedMailbox]]]

_flags_attr = FetchAttribute(b'FLAGS')
_uid_attr = FetchAttribute(b'UID')


class ConnectionState:
    """Defines the state flow of the IMAP connection. Determines if a command
    can be processed at that point in the connection, and interacts with the
    backend plugin.

    """

    def __init__(self, login: LoginProtocol, config: IMAPConfig) -> None:
        super().__init__()
        self.config = config
        self.ssl_context = config.ssl_context
        self.auth = config.initial_auth
        self.login = login
        self._session: Optional[SessionInterface] = None
        self._selected: Optional[SelectedMailbox] = None
        self._capability = list(config.initial_capability)

    @property
    def session(self) -> SessionInterface:
        if self._session is None:
            raise RuntimeError()  # State checking should prevent this.
        return self._session

    @property
    def selected(self) -> SelectedMailbox:
        if self._selected is None:
            raise RuntimeError()  # State checking should prevent this.
        return self._selected

    @property
    def capability(self) -> Capability:
        if self._session:
            return Capability(self._capability)
        else:
            if self.auth.get_server(b'PLAIN') is None:
                logindisabled = [b'LOGINDISABLED']
            else:
                logindisabled = []
            return Capability(self._capability + logindisabled +
                              [b'AUTH=%b' % mech.name for mech in
                               self.auth.server_mechanisms])

    async def do_cleanup(self) -> None:
        try:
            await self.session.cleanup()
        except Exception:
            pass

    async def do_greeting(self) -> Response:
        preauth_creds = self.config.preauth_credentials
        if preauth_creds:
            self._session = await self.login(preauth_creds, self.config)
        elif socket_info.get().from_localhost:
            self.auth = self.config.insecure_auth
        resp_cls = ResponsePreAuth if preauth_creds else ResponseOk
        return resp_cls(b'*', self.config.greeting, self.capability)

    async def do_authenticate(self, cmd: _AuthCommands,
                              creds: Optional[AuthenticationCredentials]):
        if not creds:
            return ResponseNo(cmd.tag, b'Invalid authentication mechanism.')
        self._session = await self.login(creds, self.config)
        self._capability.extend(self.config.login_capability)
        return ResponseOk(cmd.tag, b'Authentication successful.',
                          self.capability), None

    async def do_login(self, cmd: LoginCommand) -> Response:
        if b'LOGINDISABLED' in self.capability:
            raise CommandNotAllowed(b'LOGIN is disabled.')
        creds = AuthenticationCredentials(
            cmd.userid.decode('utf-8', 'surrogateescape'),
            cmd.password.decode('utf-8', 'surrogateescape'))
        return await self.do_authenticate(cmd, creds)

    async def do_starttls(self, cmd: StartTLSCommand):
        if self.ssl_context is None:
            raise ValueError('ssl_context is None')
        try:
            self._capability.remove(b'STARTTLS')
        except ValueError:
            raise CommandNotAllowed(b'STARTTLS not available.')
        self.auth = self.config.insecure_auth
        return ResponseOk(cmd.tag, b'Ready to handshake.'), None

    async def do_capability(self, cmd: CapabilityCommand):
        response = ResponseOk(cmd.tag, b'Capabilities listed.')
        response.add_untagged(Response(b'*', self.capability.string))
        return response, None

    async def do_noop(self, cmd: NoOpCommand):
        updates = None
        if self._selected and self._session:
            updates = await self.session.check_mailbox(self.selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_select(self, cmd: SelectCommand):
        self._selected = None
        mailbox, updates = await self.session.select_mailbox(
            cmd.mailbox, cmd.readonly)
        if updates.readonly:
            num_recent = mailbox.recent
            resp = ResponseOk(cmd.tag, b'Selected mailbox.',
                              ResponseCode.of(b'READ-ONLY'))
            resp.add_untagged_ok(b'Read-only mailbox.', PermanentFlags([]))
        else:
            num_recent = updates.session_flags.recent
            resp = ResponseOk(cmd.tag, b'Selected mailbox.',
                              ResponseCode.of(b'READ-WRITE'))
            resp.add_untagged_ok(b'Flags permitted.',
                                 PermanentFlags(mailbox.permanent_flags))
        messages = updates.messages
        resp.add_untagged(FlagsResponse(mailbox.flags))
        resp.add_untagged(ExistsResponse(messages.exists))
        resp.add_untagged(RecentResponse(num_recent))
        resp.add_untagged_ok(b'Predicted next UID.',
                             UidNext(mailbox.next_uid))
        resp.add_untagged_ok(b'UIDs valid.',
                             UidValidity(mailbox.uid_validity))
        if mailbox.first_unseen:
            resp.add_untagged_ok(b'First unseen message.',
                                 Unseen(mailbox.first_unseen))
        return resp, updates

    async def do_create(self, cmd: CreateCommand):
        if cmd.mailbox == 'INBOX':
            return ResponseNo(cmd.tag, b'Cannot create INBOX.'), None
        updates = await self.session.create_mailbox(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_delete(self, cmd: DeleteCommand):
        if cmd.mailbox == 'INBOX':
            return ResponseNo(cmd.tag, b'Cannot delete INBOX.'), None
        updates = await self.session.delete_mailbox(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_rename(self, cmd: RenameCommand):
        if cmd.to_mailbox == 'INBOX':
            return ResponseNo(cmd.tag, b'Cannot rename to INBOX.'), None
        updates = await self.session.rename_mailbox(
            cmd.from_mailbox, cmd.to_mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_status(self, cmd: StatusCommand):
        mailbox, updates = await self.session.get_mailbox(
            cmd.mailbox, selected=self._selected)
        data: Dict[StatusAttribute, Number] = OrderedDict()
        for attr in cmd.status_list:
            if attr == b'MESSAGES':
                data[attr] = Number(mailbox.exists)
            elif attr == b'RECENT':
                if updates and updates.guid == mailbox.guid:
                    data[attr] = Number(updates.session_flags.recent)
                else:
                    data[attr] = Number(mailbox.recent)
            elif attr == b'UNSEEN':
                data[attr] = Number(mailbox.unseen)
            elif attr == b'UIDNEXT':
                data[attr] = Number(mailbox.next_uid)
            elif attr == b'UIDVALIDITY':
                data[attr] = Number(mailbox.uid_validity)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        resp.add_untagged(StatusResponse(cmd.mailbox, data))
        return resp, updates

    async def do_append(self, cmd: AppendCommand):
        for msg in cmd.messages:
            if msg.message == b'':
                return ResponseNo(cmd.tag, b'APPEND cancelled.'), None
        append_uid, updates = await self.session.append_messages(
            cmd.mailbox, cmd.messages, selected=self._selected)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.', append_uid)
        return resp, updates

    async def do_subscribe(self, cmd: SubscribeCommand):
        updates = await self.session.subscribe(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_unsubscribe(self, cmd: UnsubscribeCommand):
        updates = await self.session.unsubscribe(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_list(self, cmd: ListCommand):
        mailboxes, updates = await self.session.list_mailboxes(
            cmd.ref_name, cmd.filter, subscribed=cmd.only_subscribed,
            selected=self._selected)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        resp_type = LSubResponse if cmd.only_subscribed else ListResponse
        for name, sep, attrs in mailboxes:
            resp.add_untagged(resp_type(name, sep, attrs))
        return resp, updates

    async def do_check(self, cmd: CheckCommand):
        updates = await self.session.check_mailbox(
            self.selected, housekeeping=True)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_close(self, cmd: CloseCommand):
        await self.session.expunge_mailbox(self.selected)
        self._selected = None
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), None

    async def do_expunge(self, cmd: ExpungeCommand):
        updates = await self.session.expunge_mailbox(
            self.selected, cmd.uid_set)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        return resp, updates

    async def do_copy(self, cmd: CopyCommand):
        copy_uid, updates = await self.session.copy_messages(
            self.selected, cmd.sequence_set, cmd.mailbox)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.', copy_uid)
        return resp, updates

    async def do_fetch(self, cmd: FetchCommand):
        if not cmd.uid:
            self.selected.hide_expunged = True
        messages, updates = await self.session.fetch_messages(
            self.selected, cmd.sequence_set, frozenset(cmd.attributes))
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        for msg_seq, msg in messages:
            if msg.expunged:
                resp.code = ResponseCode.of(b'EXPUNGEISSUED')
            msg_attrs = MessageAttributes(msg, self.selected)
            fetch_data = msg_attrs.get_all(cmd.attributes)
            resp.add_untagged(FetchResponse(msg_seq, fetch_data))
        return resp, updates

    async def do_search(self, cmd: SearchCommand):
        if not cmd.uid:
            self.selected.hide_expunged = True
        messages, updates = await self.session.search_mailbox(
            self.selected, cmd.keys)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        msg_ids: List[int] = []
        for msg_seq, msg in messages:
            if msg.expunged:
                resp.code = ResponseCode.of(b'EXPUNGEISSUED')
            if cmd.uid:
                msg_ids.append(msg.uid)
            else:
                msg_ids.append(msg_seq)
        resp.add_untagged(SearchResponse(msg_ids))
        return resp, updates

    async def do_store(self, cmd: StoreCommand):
        if not cmd.uid:
            self.selected.hide_expunged = True
        if cmd.silent:
            self.selected.silence(cmd.sequence_set, cmd.flag_set, cmd.mode)
        messages, updates = await self.session.update_flags(
            self.selected, cmd.sequence_set, cmd.flag_set, cmd.mode)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        session_flags = self.selected.session_flags
        for msg_seq, msg in messages:
            if msg.expunged:
                resp.code = ResponseCode.of(b'EXPUNGEISSUED')
            elif cmd.silent:
                continue
            flags = msg.get_flags(session_flags)
            fetch_data: List[Tuple[FetchAttribute, MaybeBytes]] = [
                (_flags_attr, ListP(flags, sort=True))]
            if cmd.uid:
                fetch_data.append((_uid_attr, Number(msg.uid)))
            resp.add_untagged(FetchResponse(msg_seq, fetch_data))
        return resp, updates

    async def do_idle(self, cmd: IdleCommand):
        if b'IDLE' not in self.capability:
            raise CommandNotAllowed(b'IDLE is disabled.')
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), None

    @classmethod
    async def do_logout(cls, cmd: LogoutCommand):
        raise CloseConnection()

    async def receive_updates(self, cmd: IdleCommand, done: Event) \
            -> Iterable[Response]:
        selected = await self.session.check_mailbox(
            self.selected, wait_on=done)
        self._selected, untagged = selected.fork(cmd)
        return untagged

    @classmethod
    def _get_bad_response(cls, cmd: InvalidCommand) -> ResponseBad:
        if not cmd.command_name:
            return ResponseBad(cmd.tag, b'Command not given.')
        elif not cmd.command_type:
            msg = b'%b: Command not implemented.' % cmd.command_name
            return ResponseBad(cmd.tag, msg)
        else:
            msg = b'%b: Invalid arguments.' % cmd.command_name
            return ResponseBad(cmd.tag, msg)

    @classmethod
    def _get_func_name(cls, cmd: Command) -> str:
        cmd_type = type(cmd)
        while cmd_type.delegate:
            cmd_type = cmd_type.delegate
        cmd_str = str(cmd_type.command, 'ascii').lower()
        return 'do_' + cmd_str

    async def do_command(self, cmd: Command):
        if isinstance(cmd, InvalidCommand):
            return self._get_bad_response(cmd)
        elif self._session and isinstance(cmd, CommandNonAuth):
            msg = cmd.command + b': Already authenticated.'
            return ResponseBad(cmd.tag, msg)
        elif not self._session and isinstance(cmd, CommandAuth):
            msg = cmd.command + b': Must authenticate first.'
            return ResponseBad(cmd.tag, msg)
        elif not self._selected and isinstance(cmd, CommandSelect):
            msg = cmd.command + b': Must select a mailbox first.'
            return ResponseBad(cmd.tag, msg)
        func_name = self._get_func_name(cmd)
        try:
            func: _CommandFunc = getattr(self, func_name)
        except AttributeError:
            return ResponseNo(cmd.tag, cmd.command + b': Not Implemented')
        response, selected = await func(cmd)
        if selected:
            self._selected, untagged = selected.fork(cmd)
            response.add_untagged(*untagged)
        return response




from typing import Optional, Tuple

from pymap.exceptions import AppendFailure
from pymap.filter import FilterCompiler
from pymap.interfaces.filter import FilterInterface
from pymap.interfaces.session import AppendMessage
from sievelib.parser import Parser  # type: ignore

from .runner import SieveRunner
from .util import unquote

__all__ = ['SieveParseError' 'SieveCompiler', 'SieveFilter']


class SieveParseError(ValueError):
    """Parsing the sieve script failed."""
    pass


class SieveCompiler(FilterCompiler[bytes]):
    """Compiles sieve scripts into :class:`SieveFilter` objects."""

    def __init__(self) -> None:
        super().__init__()
        self.parser = Parser()

    async def compile(self, value: bytes) -> 'SieveFilter':
        parser = self.parser
        try:
            success = parser.parse(value)
        except Exception as exc:
            raise SieveParseError('Unhandled parsing error') from exc
        if success:
            runner = SieveRunner(parser.result)
            return SieveFilter(runner)
        else:
            raise SieveParseError(parser.error)


class SieveFilter(FilterInterface):
    """Filter implementation for applying sieve scripts to the appended
    message. The sieve script applies various conditions and controls to
    ultimately produce a set of actions to apply to the message.

    Args:
        runner: The sieve script runner.

    """

    def __init__(self, runner: SieveRunner) -> None:
        super().__init__()
        self.runner = runner

    async def apply(self, sender: str, recipient: str, mailbox: str,
                    append_msg: AppendMessage) \
            -> Tuple[Optional[str], AppendMessage]:
        for action in self.runner.get_actions(sender, recipient, append_msg):
            if action.name == 'keep':
                pass
            elif action.name == 'fileinto':
                mailbox = unquote(action.arguments['mailbox'])
                return mailbox, append_msg
            elif action.name == 'reject':
                msg = unquote(action.arguments['text']).encode('utf-8')
                raise AppendFailure(mailbox, msg)
            elif action.name == 'discard':
                return None, append_msg
            else:
                msg = b'Unrecognized sieve action: %b' \
                    % action.name.encode('ascii')
                raise AppendFailure(mailbox, msg) from KeyError(action.name)
        return mailbox, append_msg




from email.headerregistry import Address, AddressHeader, SingleAddressHeader, \
    UnstructuredHeader, DateHeader, ContentDispositionHeader, \
    ContentTransferEncodingHeader
from itertools import chain
from typing import Any, SupportsBytes, Optional, Mapping, Sequence, Union, List

from ..primitives import ListP, Nil, Number, String
from ..specials import DateTime
from ...bytes import Writeable, WriteStream

__all__ = ['EnvelopeStructure', 'BodyStructure', 'MultipartBodyStructure',
           'ContentBodyStructure', 'TextBodyStructure', 'MessageBodyStructure']

_CTEHeader = ContentTransferEncodingHeader


class _Concatenated(Writeable):

    def __init__(self, parts: Sequence[SupportsBytes]) -> None:
        super().__init__()
        self.parts = parts

    def write(self, writer: WriteStream) -> None:
        for part in self.parts:
            writer.write(bytes(part))

    def __bytes__(self) -> bytes:
        return b''.join([bytes(part) for part in self.parts])


class _AddressList(Writeable):

    def __init__(self, headers: Optional[Sequence[AddressHeader]]) -> None:
        super().__init__()
        self.headers: Sequence[AddressHeader] = headers or []

    @classmethod
    def _parse(cls, address: Address):
        realname = String.build(address.display_name)
        localpart = String.build(address.username)
        domain = String.build(address.domain)
        return ListP([realname, Nil(), localpart, domain])

    @property
    def _value(self) -> Writeable:
        if self.headers:
            addresses: List[Address] = []
            for header in self.headers:
                if isinstance(header, SingleAddressHeader):
                    addresses.append(header.address)
                else:
                    addresses.extend(header.addresses)
            return ListP([self._parse(address)
                          for address in addresses])
        else:
            return Nil()

    def write(self, writer: WriteStream) -> None:
        self._value.write(writer)

    def __bytes__(self) -> bytes:
        return bytes(self._value)


class _ParamsList(Writeable):

    def __init__(self, params: Optional[Mapping[str, Any]]) -> None:
        super().__init__()
        self.params = params

    @property
    def _value(self) -> Writeable:
        if self.params:
            values = [(String.build(key), String.build(value))
                      for key, value in self.params.items()]
            return ListP(chain.from_iterable(values))
        else:
            return Nil()

    def write(self, writer: WriteStream) -> None:
        self._value.write(writer)

    def __bytes__(self) -> bytes:
        return bytes(self._value)


class EnvelopeStructure(Writeable):
    """Builds the response to an `RFC 3501 7.4.2
    <https://tools.ietf.org/html/rfc3501#section-7.4.2>`_ FETCH ENVELOPE
    request.

    Args:
        date: Original date of the message.
        subject: ``Subject:`` header.
        from_: ``From:`` headers.
        sender: ``Sender:`` headers.
        reply_to: ``Reply-To:`` headers.
        to: ``To:`` headers.
        cc: ``Cc:`` headers.
        bcc: ``Bcc`` headers.
        in_reply_to: ``In-Reply-To:`` header.
        message_id: ``Message-Id:`` header.

    """

    def __init__(self, date: Optional[DateHeader],
                 subject: Optional[UnstructuredHeader],
                 from_: Optional[Sequence[AddressHeader]],
                 sender: Optional[Sequence[SingleAddressHeader]],
                 reply_to: Optional[Sequence[AddressHeader]],
                 to: Optional[Sequence[AddressHeader]],
                 cc: Optional[Sequence[AddressHeader]],
                 bcc: Optional[Sequence[AddressHeader]],
                 in_reply_to: Optional[UnstructuredHeader],
                 message_id: Optional[UnstructuredHeader]) -> None:
        super().__init__()
        self.date = date
        self.subject = subject
        self.from_ = from_
        self.sender = sender
        self.reply_to = reply_to
        self.to = to
        self.cc = cc
        self.bcc = bcc
        self.in_reply_to = in_reply_to
        self.message_id = message_id

    @classmethod
    def empty(cls) -> 'EnvelopeStructure':
        """Return an empty envelope structure object.

        See Also:
            `RFC 2180 4.1.3
            <https://tools.ietf.org/html/rfc2180#section-4.1.3>`_

        """
        return _EmptyEnvelopeStructure()

    def _addresses(self, headers: Optional[Sequence[AddressHeader]],
                   fallback: Optional[Sequence[AddressHeader]] = None) \
            -> SupportsBytes:
        if not headers and fallback:
            return self._addresses(fallback)
        return _AddressList(headers)

    @property
    def _value(self) -> Writeable:
        datetime: Union[DateTime, Nil] = \
            DateTime(self.date.datetime) if self.date else Nil()
        return ListP([datetime,
                      String.build(self.subject),
                      self._addresses(self.from_),
                      self._addresses(self.sender, self.from_),
                      self._addresses(self.reply_to, self.from_),
                      self._addresses(self.to),
                      self._addresses(self.cc),
                      self._addresses(self.bcc),
                      String.build(self.in_reply_to),
                      String.build(self.message_id)])

    def write(self, writer: WriteStream) -> None:
        self._value.write(writer)

    def __bytes__(self) -> bytes:
        return bytes(self._value)


class BodyStructure(Writeable):
    """Parent class for the response to an `RFC 3501 7.4.2
    <https://tools.ietf.org/html/rfc3501#section-7.4.2>`_ FETCH BODYSTRUCTURE
    request. This class should not be used directly.

    Args:
        maintype: Main-type of the ``Content-Type:`` header.
        subtype: Sub-type of the ``Content-Type:`` header.
        content_type_params: Parameters from the ``Content-Type:`` header.
        content_disposition: ``Content-Disposition:`` header.
        content_language: ``Content-Language:`` header.
        content_location: ``Content-Location:`` header.

    """

    def __init__(self, maintype: str, subtype: str,
                 content_type_params: Optional[Mapping[str, Any]],
                 content_disposition: Optional[ContentDispositionHeader],
                 content_language: Optional[UnstructuredHeader],
                 content_location: Optional[UnstructuredHeader]) -> None:
        super().__init__()
        self.maintype = maintype
        self.subtype = subtype
        self.content_type_params = content_type_params
        self.content_disposition = content_disposition
        self.content_language = content_language
        self.content_location = content_location

    @classmethod
    def empty(cls) -> 'BodyStructure':
        """Return an empty body structure object.

        See Also:
            `RFC 2180 4.1.3
            <https://tools.ietf.org/html/rfc2180#section-4.1.3>`_

        """
        return _EmptyBodyStructure()

    @property
    def _value(self) -> ListP:
        raise NotImplementedError

    def write(self, writer: WriteStream) -> None:
        self._value.write(writer)

    def __bytes__(self) -> bytes:
        return bytes(self._value)

    @property
    def extended(self) -> ListP:
        """The body structure attributes with extension data."""
        raise NotImplementedError


class MultipartBodyStructure(BodyStructure):
    """:class:`BodyStructure` sub-class for ``multipart/*`` messages. The
    response is made up of the BODYSTRUCTUREs of all sub-parts.

    Args:
        subtype: Sub-type of the ``Content-Type:`` header.
        content_type_params: Parameters from the ``Content-Type:`` header.
        content_disposition: ``Content-Disposition:`` header.
        content_language: ``Content-Language:`` header.
        content_location: ``Content-Location:`` header.
        parts: Sub-part body structure objects.

    """

    def __init__(self, subtype: str,
                 content_type_params: Optional[Mapping[str, Any]],
                 content_disposition: Optional[ContentDispositionHeader],
                 content_language: Optional[UnstructuredHeader],
                 content_location: Optional[UnstructuredHeader],
                 parts: Sequence[BodyStructure]) -> None:
        super().__init__('multipart', subtype, content_type_params,
                         content_disposition, content_language,
                         content_location)
        self.parts = parts

    @property
    def _value(self) -> ListP:
        return ListP([_Concatenated(self.parts), String.build(self.subtype)])

    @property
    def extended(self) -> ListP:
        """The body structure attributes with extension data."""
        parts = [part.extended for part in self.parts]
        return ListP([_Concatenated(parts), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_disposition),
                      String.build(self.content_language),
                      String.build(self.content_location)])


class ContentBodyStructure(BodyStructure):
    """:class:`BodyStructure` sub-class for any content type that does not
    fit the other sub-classes.

    Args:
        maintype: Main-type of the ``Content-Type:`` header.
        subtype: Sub-type of the ``Content-Type:`` header.
        content_type_params: Parameters from the ``Content-Type:`` header.
        content_disposition: ``Content-Disposition:`` header.
        content_language: ``Content-Language:`` header.
        content_location: ``Content-Location:`` header.
        content_id: ``Content-Id:`` header.
        content_description: ``Content-Description:`` header.
        content_transfer_encoding: ``Content-Transfer-Encoding:`` header.
        body_md5: MD5 hash of the body content.
        size: Size of the message body, in bytes.

    """

    def __init__(self, maintype: str, subtype: str,
                 content_type_params: Optional[Mapping[str, Any]],
                 content_disposition: Optional[ContentDispositionHeader],
                 content_language: Optional[UnstructuredHeader],
                 content_location: Optional[UnstructuredHeader],
                 content_id: Optional[UnstructuredHeader],
                 content_description: Optional[UnstructuredHeader],
                 content_transfer_encoding: Optional[_CTEHeader],
                 body_md5: Optional[str],
                 size: int) -> None:
        super().__init__(maintype, subtype, content_type_params,
                         content_disposition, content_language,
                         content_location)
        self.content_id = content_id
        self.content_description = content_description
        self.content_transfer_encoding = content_transfer_encoding
        self.body_md5 = body_md5
        self.size = size

    @property
    def _value(self) -> ListP:
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding,
                                   fallback=b'7BIT'),
                      Number(self.size)])

    @property
    def extended(self) -> ListP:
        """The body structure attributes with extension data."""
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding,
                                   fallback=b'7BIT'),
                      Number(self.size),
                      String.build(self.body_md5),
                      String.build(self.content_disposition),
                      String.build(self.content_language),
                      String.build(self.content_location)])


class TextBodyStructure(ContentBodyStructure):
    """:class:`BodyStructure` sub-class for ``text/*`` messages.

    Args:
        subtype: Sub-type of the ``Content-Type:`` header.
        content_type_params: Parameters from the ``Content-Type:`` header.
        content_disposition: ``Content-Disposition:`` header.
        content_language: ``Content-Language:`` header.
        content_location: ``Content-Location:`` header.
        content_id: ``Content-Id:`` header.
        content_description: ``Content-Description:`` header.
        content_transfer_encoding: ``Content-Transfer-Encoding:`` header.
        body_md5: MD5 hash of the body content.
        size: Size of the message body, in bytes.
        lines: Length of the message body, in lines.

    """

    def __init__(self, subtype: str,
                 content_type_params: Optional[Mapping[str, Any]],
                 content_disposition: Optional[ContentDispositionHeader],
                 content_language: Optional[UnstructuredHeader],
                 content_location: Optional[UnstructuredHeader],
                 content_id: Optional[UnstructuredHeader],
                 content_description: Optional[UnstructuredHeader],
                 content_transfer_encoding: Optional[_CTEHeader],
                 body_md5: Optional[str],
                 size: int, lines: int) -> None:
        super().__init__('text', subtype, content_type_params,
                         content_disposition, content_language,
                         content_location, content_id, content_description,
                         content_transfer_encoding, body_md5, size)
        self.lines = lines

    @property
    def _value(self) -> ListP:
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding,
                                   fallback=b'7BIT'),
                      Number(self.size), Number(self.lines)])

    @property
    def extended(self) -> ListP:
        """The body structure attributes with extension data."""
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding,
                                   fallback=b'7BIT'),
                      Number(self.size), Number(self.lines),
                      String.build(self.body_md5),
                      String.build(self.content_disposition),
                      String.build(self.content_language),
                      String.build(self.content_location)])


class MessageBodyStructure(ContentBodyStructure):
    """:class:`BodyStructure` sub-class for ``message/rfc822`` messages.

    Args:
        content_type_params: Parameters from the ``Content-Type:`` header.
        content_disposition: ``Content-Disposition:`` header.
        content_language: ``Content-Language:`` header.
        content_location: ``Content-Location:`` header.
        content_id: ``Content-Id:`` header.
        content_description: ``Content-Description:`` header.
        content_transfer_encoding: ``Content-Transfer-Encoding:`` header.
        body_md5: MD5 hash of the body content.
        size: Size of the message body, in bytes.
        lines: Length of the message body, in lines.
        envelope_structure: Contained message's envelope structure.
        body_structure: Contained message's body structure.

    """

    def __init__(self, content_type_params: Optional[Mapping[str, Any]],
                 content_disposition: Optional[ContentDispositionHeader],
                 content_language: Optional[UnstructuredHeader],
                 content_location: Optional[UnstructuredHeader],
                 content_id: Optional[UnstructuredHeader],
                 content_description: Optional[UnstructuredHeader],
                 content_transfer_encoding: Optional[_CTEHeader],
                 body_md5: Optional[str],
                 size: int, lines: int,
                 envelope_structure: EnvelopeStructure,
                 body_structure: BodyStructure) -> None:
        super().__init__('message', 'rfc822', content_type_params,
                         content_disposition, content_language,
                         content_location, content_id, content_description,
                         content_transfer_encoding, body_md5, size)
        self.lines = lines
        self.envelope_structure = envelope_structure
        self.body_structure = body_structure

    @property
    def _value(self) -> ListP:
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding,
                                   fallback=b'7BIT'),
                      Number(self.size),
                      self.envelope_structure,
                      self.body_structure,
                      Number(self.lines)])

    @property
    def extended(self) -> ListP:
        """The body structure attributes with extension data."""
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding,
                                   fallback=b'7BIT'),
                      Number(self.size),
                      self.envelope_structure,
                      self.body_structure.extended,
                      Number(self.lines),
                      String.build(self.body_md5),
                      String.build(self.content_disposition),
                      String.build(self.content_language),
                      String.build(self.content_location)])


class _EmptyEnvelopeStructure(EnvelopeStructure):

    def __init__(self) -> None:
        super().__init__(None, None, None, None, None, None, None, None,
                         None, None)


class _EmptyBodyStructure(TextBodyStructure):

    def __init__(self) -> None:
        super().__init__('plain', None, None, None, None, None,
                         None, None, None, 0, 0)




import asyncio
import os
import os.path
import tempfile
from argparse import ArgumentParser, Namespace
from asyncio import Task, AbstractServer, CancelledError
from typing import Optional

from grpclib.server import Server  # type: ignore
from pymap.config import IMAPConfig
from pymap.interfaces.backend import BackendInterface, ServiceInterface

from .handlers import AdminHandlers

__all__ = ['AdminService']


class AdminService(ServiceInterface):
    """A pymap service implemented using a `grpc <https://grpc.io/>`_ server
    to perform admin functions on a running backend.

    """

    def __init__(self, path: str, server: AbstractServer) -> None:
        super().__init__()
        self._path = path
        self._server = server
        self._task = asyncio.create_task(self._run())

    @classmethod
    def get_socket_path(cls) -> str:
        """Return the prioritized list of locations to create or check for the
        UNIX socket file listening for admin requests.

        """
        return os.path.join(tempfile.gettempdir(), 'pymap-admin.sock')

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        group = parser.add_argument_group('admin service')
        group.add_argument('--admin-socket', metavar='PATH', dest='admin_sock',
                           help='path to socket file')
        group.add_argument('--no-filter', action='store_true',
                           help='do not filter appended messages')

    @classmethod
    async def start(cls, backend: BackendInterface,
                    config: IMAPConfig) -> 'AdminService':
        config = backend.config
        path: Optional[str] = config.args.admin_sock
        handlers = AdminHandlers(backend)
        server = Server([handlers], loop=asyncio.get_event_loop())
        path = await cls._start(path, server)
        cls._chown(path, config.args)
        return cls(path, server)

    @classmethod
    async def _start(cls, path: Optional[str], server: Server) -> str:
        if path:
            await server.start(path=path)
            return path
        else:
            path = cls.get_socket_path()
            await server.start(path=path)
            return path

    @classmethod
    def _chown(cls, path: str, args: Namespace) -> None:
        uid = args.set_uid or -1
        gid = args.set_gid or -1
        if uid >= 0 or gid >= 0:
            os.chown(path, uid, gid)

    @property
    def task(self) -> Task:
        return self._task

    def _unlink_path(self) -> None:
        try:
            os.unlink(self._path)
        except OSError:
            pass

    async def _run(self) -> None:
        server = self._server
        try:
            await server.wait_closed()
        except CancelledError:
            server.close()
            await server.wait_closed()
        finally:
            self._unlink_path()
